"""
Module to provide processing for the non-leaf scenarios that may contain container blocks.
"""

from __future__ import annotations

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_processor import BlockQuoteProcessor
from pymarkdown.container_blocks.container_block_nested_processor import (
    ContainerBlockNestedProcessor,
)
from pymarkdown.container_blocks.container_grab_bag import ContainerGrabBag
from pymarkdown.container_blocks.container_indices import ContainerIndices
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.leaf_blocks.leaf_block_processor_paragraph import (
    LeafBlockProcessorParagraph,
)
from pymarkdown.list_blocks.list_block_processor import ListBlockProcessor
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class ContainerBlockNonLeafProcessor:
    """
    Class to provide processing for the non-leaf scenarios that may contain container blocks.
    """

    @staticmethod
    def handle_non_leaf_block(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> None:
        # POGGER.debug("ttp>>:$:<", position_marker.text_to_parse)
        # POGGER.debug("index_number>>:$:<", position_marker.index_number)
        # POGGER.debug("index_indent>>:$:<", position_marker.index_indent)
        ContainerBlockNonLeafProcessor.__handle_pre_processed_indent(
            parser_state,
            position_marker,
            grab_bag,
        )
        if ContainerBlockNonLeafProcessor.__look_for_override(
            parser_state,
            position_marker,
            grab_bag,
        ):
            ContainerBlockNonLeafProcessor.__handle_normal_containers(
                parser_state,
                position_marker,
                grab_bag,
            )

    @staticmethod
    def __handle_pre_processed_indent(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> None:
        assert (
            grab_bag.extracted_whitespace is not None
        ), "Extracted whitespace must be defined at this point."
        need_leading_whitespace_processing = (
            grab_bag.container_depth < (len(parser_state.token_stack) - 1)
            and len(grab_bag.extracted_whitespace) >= 4
            and not parser_state.token_stack[-1].is_html_block
            and not parser_state.token_stack[-1].is_fenced_code_block
        )
        need_trailing_indent_processing = (
            grab_bag.container_depth >= len(parser_state.token_stack) - 1
            and position_marker.index_number == -1
        )
        if need_trailing_indent_processing:
            POGGER.debug(">>trailing_indent_processing")
            ContainerBlockNonLeafProcessor.__handle_trailing_indent_with_block_quote(
                parser_state, grab_bag
            )
        elif need_leading_whitespace_processing:
            POGGER.debug(">>leading_whitespace_preprocessing")
            ContainerBlockNonLeafProcessor.__determine_leading_whitespace_preprocessing(
                parser_state,
                position_marker,
                grab_bag,
            )

    @staticmethod
    def __look_for_override(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> bool:
        if grab_bag.have_pre_processed_indent:
            (
                grab_bag.can_continue,
                grab_bag.line_to_parse,
                grab_bag.start_index,
                grab_bag.removed_chars_at_start_of_line,
            ) = (
                True,
                position_marker.text_to_parse[grab_bag.indent_already_processed :],
                grab_bag.indent_already_processed,
                grab_bag.indent_already_processed,
            )
        keep_processing = (
            not grab_bag.can_continue
            and not grab_bag.do_skip_containers_before_leaf_blocks
        )
        if keep_processing:
            last_container_stack_token = parser_state.token_stack[
                parser_state.find_last_container_on_stack()
            ]
            POGGER.debug("last_container_stack_token>>$", last_container_stack_token)
            if last_container_stack_token.is_block_quote:
                block_token = cast(
                    BlockQuoteMarkdownToken,
                    last_container_stack_token.matching_markdown_token,
                )
                assert (
                    block_token.bleading_spaces is not None
                ), "Bleading spaces must be defined by this point."
                split_spaces = block_token.bleading_spaces.split(
                    ParserHelper.newline_character
                )
                last_leading_space = split_spaces[-1]
                ex_ws_index, _ = ParserHelper.extract_spaces(last_leading_space, 0)
                assert (
                    grab_bag.adj_ws is not None
                ), "Adjusted whitespace must be defined by this point."
                if (
                    len(grab_bag.adj_ws) >= 4
                    and grab_bag.indent_used_by_container
                    and grab_bag.indent_used_by_container == ex_ws_index
                ):
                    keep_processing = False
                    grab_bag.indent_already_processed = len(last_leading_space)
                    grab_bag.weird_adjusted_text = last_leading_space
                    (
                        grab_bag.can_continue,
                        grab_bag.line_to_parse,
                        grab_bag.start_index,
                        grab_bag.removed_chars_at_start_of_line,
                    ) = (
                        True,
                        position_marker.text_to_parse[
                            grab_bag.indent_already_processed :
                        ],
                        grab_bag.indent_already_processed,
                        grab_bag.indent_already_processed,
                    )
        return keep_processing

    @staticmethod
    def __handle_normal_containers(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> None:
        (
            did_process,
            avoid_block_starts,
        ) = ContainerBlockNonLeafProcessor.__check_for_container_starts(
            parser_state,
            position_marker,
            grab_bag,
        )

        if grab_bag.requeue_line_info or grab_bag.did_blank:
            grab_bag.can_continue = False
            # if grab_bag.requeue_line_info:
            #     grab_bag.extend_container_tokens_with_leaf_tokens()
            return

        nested_force_list_continuation = (
            ContainerBlockNonLeafProcessor.__handle_nested_blocks(
                parser_state,
                position_marker,
                did_process,
                avoid_block_starts,
                grab_bag,
            )
        )
        if grab_bag.do_force_list_continuation:
            POGGER.debug(
                "nested_force_list_continuation=$", nested_force_list_continuation
            )
            assert (
                not nested_force_list_continuation
            ), "This cannot be a nested and forced list continuation."
        if grab_bag.can_continue:
            did_process = ContainerBlockNonLeafProcessor.__handle_block_continuations(
                parser_state,
                position_marker,
                did_process,
                grab_bag,
            )

    @staticmethod
    def __handle_trailing_indent_with_block_quote(
        parser_state: ParserState,
        grab_bag: ContainerGrabBag,
    ) -> None:
        grab_bag.did_indent_processing = True
        assert (
            grab_bag.extracted_whitespace is not None
        ), "Extracted whitespace must be defined at this point."
        for stack_capture_index in range(1, len(parser_state.token_stack)):
            POGGER.debug(
                "$>stack:$:",
                stack_capture_index,
                parser_state.token_stack[stack_capture_index],
            )
            inner_token = parser_state.token_stack[
                stack_capture_index
            ].matching_markdown_token
            assert (
                inner_token is not None
            ), "All container and leaf tokens have a matching markdown token."
            POGGER.debug(
                "$>token:$:",
                stack_capture_index,
                inner_token,
            )
            if inner_token.is_block_quote_start:
                block_quote_token = cast(BlockQuoteMarkdownToken, inner_token)
                assert (
                    block_quote_token.bleading_spaces is not None
                ), "Bleading spaces must be defined by this point."
                split_spaces = block_quote_token.bleading_spaces.split(
                    ParserHelper.newline_character
                )
                grab_bag.indent_already_processed = len(split_spaces[-1])
            else:
                assert (
                    inner_token.is_list_start
                ), "If it is not a block start, it must be a list start."
                list_token = cast(ListStartMarkdownToken, inner_token)
                grab_bag.indent_already_processed = list_token.indent_level
        delta = len(grab_bag.extracted_whitespace) - grab_bag.indent_already_processed
        POGGER.debug("len(ws)=$", len(grab_bag.extracted_whitespace))
        POGGER.debug("len(containers)=$", grab_bag.indent_already_processed)
        grab_bag.have_pre_processed_indent = (
            grab_bag.indent_already_processed != -1 and delta >= 4
        )

        grab_bag.do_skip_containers_before_leaf_blocks = (
            grab_bag.have_pre_processed_indent
        )

    @staticmethod
    def __determine_leading_whitespace_preprocessing(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> None:
        assert (
            grab_bag.extracted_whitespace is not None
        ), "Extracted whitespace must be defined at this point."
        grab_bag.indent_used_by_container = 0
        if grab_bag.is_para_continue and not grab_bag.container_depth:
            ContainerBlockNonLeafProcessor.__special_list_block_block(
                parser_state, position_marker, grab_bag
            )
        elif not grab_bag.is_para_continue:
            POGGER.debug(
                "position_marker.index_number:$:", position_marker.index_number
            )
            if position_marker.index_number == -1 and grab_bag.container_depth:
                leading_whitespace = (
                    ContainerBlockNonLeafProcessor.__calculate_indent_used_by_container(
                        parser_state,
                        position_marker,
                        grab_bag.extracted_whitespace,
                        grab_bag,
                    )
                )
            else:
                assert (
                    grab_bag.extracted_whitespace is not None
                ), "Extracted whitespace must be defined by this point."
                leading_whitespace = grab_bag.extracted_whitespace

            if len(leading_whitespace) >= 4:
                POGGER.debug(">>leading_whitespace_processing")
                ContainerBlockNonLeafProcessor.__handle_leading_whitespace(
                    parser_state,
                    position_marker,
                    leading_whitespace,
                    grab_bag,
                )
                (
                    grab_bag.did_indent_processing,
                    grab_bag.was_indent_already_processed,
                ) = (True, bool(grab_bag.container_depth))
            POGGER.debug("parser_state.token_document=$", parser_state.token_document)

    @staticmethod
    def __check_for_container_starts(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[bool, bool]:
        parser_state.nested_list_start = None
        ulist_index = -1
        olist_index = -1
        (
            did_process,
            block_index,
            avoid_block_starts,
        ) = ContainerBlockNonLeafProcessor.__get_block_start_index(
            parser_state,
            position_marker,
            grab_bag,
        )
        if grab_bag.can_continue:
            # POGGER.debug(">>avoid_block_starts>>$", avoid_block_starts)
            # POGGER.debug(">>did_process>>$", did_process)

            (
                did_process,
                ulist_index,
            ) = ContainerBlockNonLeafProcessor.__get_list_start_index(
                parser_state,
                position_marker,
                True,
                did_process,
                grab_bag,
            )

        if grab_bag.can_continue:
            # POGGER.debug("was_container_start>>$", was_container_start)
            (
                did_process,
                olist_index,
            ) = ContainerBlockNonLeafProcessor.__get_list_start_index(
                parser_state,
                position_marker,
                False,
                did_process,
                grab_bag,
            )

        grab_bag.end_container_indices = ContainerIndices(
            block_index=block_index, ulist_index=ulist_index, olist_index=olist_index
        )
        return (
            did_process,
            avoid_block_starts,
        )

    @staticmethod
    def __handle_nested_blocks(
        parser_state: ParserState,
        position_marker: PositionMarker,
        was_container_start: bool,
        avoid_block_starts: bool,
        grab_bag: ContainerGrabBag,
    ) -> bool:
        grab_bag.last_list_start_index = 0
        if grab_bag.end_container_indices.block_index != -1:
            assert grab_bag.last_block_quote_index in (
                grab_bag.end_container_indices.block_index - 1,
                grab_bag.end_container_indices.block_index,
            ), "Sanity check for block index failed."
        elif grab_bag.end_container_indices.olist_index != -1:
            grab_bag.last_list_start_index = grab_bag.end_container_indices.olist_index
        elif grab_bag.end_container_indices.ulist_index != -1:
            grab_bag.last_list_start_index = grab_bag.end_container_indices.ulist_index

        if not parser_state.token_stack[-1].is_fenced_code_block:
            new_position_marker = PositionMarker(
                position_marker.line_number,
                grab_bag.start_index,
                grab_bag.line_to_parse,
            )
            # POGGER.debug("was_container_start>>$", was_container_start)
            (
                nested_removed_text,
                was_indent_text_added,
                nested_force_list_continuation,
            ) = ContainerBlockNestedProcessor.handle_nested_container_blocks(
                parser_state,
                new_position_marker,
                was_container_start,
                avoid_block_starts,
                grab_bag,
            )
        else:
            (
                nested_removed_text,
                was_indent_text_added,
                nested_force_list_continuation,
            ) = (None, False, False)

        POGGER.debug_with_visible_whitespace(
            "nested_removed_text>>>:$:", nested_removed_text
        )
        if nested_removed_text is not None:
            grab_bag.text_removed_by_container = nested_removed_text
        grab_bag.can_continue = not (
            grab_bag.container_depth or grab_bag.did_blank or was_indent_text_added
        )
        POGGER.debug(
            ">>can_continue>>:$: = not(container_depth($) or "
            + "did_process_blank_line($) or was_indent_text_added($))",
            grab_bag.can_continue,
            grab_bag.container_depth,
            grab_bag.did_blank,
            was_indent_text_added,
        )
        return nested_force_list_continuation

    @staticmethod
    def __handle_block_continuations(
        parser_state: ParserState,
        position_marker: PositionMarker,
        did_process: bool,
        grab_bag: ContainerGrabBag,
    ) -> bool:
        if not did_process:
            did_process = ContainerBlockNonLeafProcessor.__process_list_in_progress(
                parser_state,
                position_marker.index_indent,
                grab_bag,
            )
            if not grab_bag.requeue_line_info:
                ContainerBlockNonLeafProcessor.__process_lazy_lines(
                    parser_state,
                    position_marker,
                    grab_bag,
                )
        else:
            is_paragraph_continuation = (
                bool(parser_state.token_stack)
                and parser_state.token_stack[-1].is_paragraph
            )
            list_index = parser_state.find_last_list_block_on_stack()
            block_index = parser_state.find_last_block_quote_on_stack()
            assert (
                not grab_bag.was_paragraph_continuation
            ), "This cannot be a paragraph continuation."
            grab_bag.was_paragraph_continuation = (
                is_paragraph_continuation and block_index > list_index
            )

        grab_bag.can_continue = not grab_bag.requeue_line_info
        return did_process

    @staticmethod
    def __special_list_block_block(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> None:
        if not (
            parser_state.token_stack[1].is_list
            and parser_state.token_stack[2].is_block_quote
            and parser_state.token_stack[3].is_block_quote
            and parser_state.token_stack[4].is_paragraph
            and parser_state.token_stack[1].matching_markdown_token
            and parser_state.token_stack[3].matching_markdown_token
            and parser_state.token_stack[1].matching_markdown_token.line_number
            == parser_state.token_stack[3].matching_markdown_token.line_number
        ):
            return
        list_token = cast(
            ListStartMarkdownToken,
            parser_state.token_stack[1].matching_markdown_token,
        )
        container_x_used_indent = list_token.indent_level
        assert (
            grab_bag.extracted_whitespace is not None
        ), "Extracted whitespace must be defined at this point."
        extracted_whitespace_length = len(grab_bag.extracted_whitespace)
        POGGER.debug("text_to_parse=:$:", position_marker.text_to_parse)
        POGGER.debug(
            "text_to_parse[rt($)]=:$:",
            extracted_whitespace_length,
            position_marker.text_to_parse[extracted_whitespace_length],
        )
        POGGER.debug("index_number=:$:", position_marker.index_number)
        if not (
            container_x_used_indent != extracted_whitespace_length
            and position_marker.text_to_parse[extracted_whitespace_length] == ">"
            and extracted_whitespace_length - container_x_used_indent >= 4
        ):
            return
        POGGER.debug("__special_list_block_block>>list_token>>$", list_token)
        list_token.add_leading_spaces(
            grab_bag.extracted_whitespace[:container_x_used_indent]
        )
        POGGER.debug("__special_list_block_block>>list_token>>$", list_token)
        (
            grab_bag.do_skip_containers_before_leaf_blocks,
            grab_bag.did_indent_processing,
            grab_bag.have_pre_processed_indent,
            grab_bag.do_force_leaf_token_parse,
            grab_bag.indent_already_processed,
        ) = (True, True, True, True, container_x_used_indent)

    @staticmethod
    def __calculate_indent_used_by_container(
        parser_state: ParserState,
        position_marker: PositionMarker,
        leading_whitespace: Optional[str],
        grab_bag: ContainerGrabBag,
    ) -> str:
        POGGER.debug("original_line_to_parse:$:", parser_state.original_line_to_parse)
        POGGER.debug("leading_whitespace:$:", leading_whitespace)
        POGGER.debug("text_to_parse=:$:", position_marker.text_to_parse)
        POGGER.debug("parser_state.token_stack=:$:", parser_state.token_stack)
        stack_search_index, current_indent = 1, 0
        assert (
            parser_state.original_line_to_parse is not None
        ), "Original line must be defined by this point."
        while stack_search_index <= grab_bag.container_depth:
            if parser_state.token_stack[stack_search_index].is_block_quote:
                current_indent = parser_state.original_line_to_parse.find(">")
                assert (
                    current_indent != -1
                ), "Since we have a block quote, there should be at least one block quote start character."
                assert (
                    parser_state.original_line_to_parse[current_indent + 1]
                    == ParserHelper.space_character
                ), "After the start character must be a space."
                current_indent += 1

                # TODO add tests with no space between `>` and next block
                assert (
                    current_indent < len(parser_state.original_line_to_parse)
                    and parser_state.original_line_to_parse[current_indent]
                    == ParserHelper.space_character
                ), "After the start character must be a space."
                current_indent += 1
            else:
                assert parser_state.token_stack[
                    stack_search_index
                ].is_list, "If not a block quote, this must be a list."
                list_token = cast(
                    ListStartMarkdownToken, parser_state.token_stack[stack_search_index]
                )
                delta = list_token.indent_level - current_indent
                POGGER.debug("delta=:$:", delta)
                current_indent += delta
            stack_search_index += 1
        assert (
            stack_search_index > grab_bag.container_depth
        ), "The new index must be more than the container depth."
        _, leading_whitespace = ParserHelper.extract_spaces_verified(
            parser_state.original_line_to_parse, current_indent
        )
        POGGER.debug("leading_whitespace=:$:", leading_whitespace)
        grab_bag.indent_used_by_container = current_indent
        return leading_whitespace

    @staticmethod
    def __handle_leading_whitespace(
        parser_state: ParserState,
        position_marker: PositionMarker,
        leading_whitespace: str,
        grab_bag: ContainerGrabBag,
    ) -> None:
        grab_bag.indent_already_processed, found_stack_index, remaining_whitespace = (
            0,
            0,
            leading_whitespace[:],
        )
        POGGER.debug(">remaining_whitespace:$:", remaining_whitespace)
        for i in range(grab_bag.container_depth + 1, len(parser_state.token_stack)):
            (
                do_break,
                found_stack_index,
                remaining_whitespace,
            ) = ContainerBlockNonLeafProcessor.__handle_leading_whitespace_loop(
                parser_state,
                i,
                remaining_whitespace,
                grab_bag,
            )
            if do_break:
                break
        POGGER.debug(">stack_index:$:", found_stack_index)
        if found_stack_index:
            grab_bag.have_pre_processed_indent = True
            close_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                include_lists=True,
                include_block_quotes=True,
                was_forced=True,
                until_this_index=found_stack_index,
            )
            grab_bag.extend_container_tokens(close_tokens)
            ind = parser_state.find_last_container_on_stack()
            if parser_state.token_stack[ind].is_list:
                ind2 = parser_state.find_last_block_quote_on_stack()
                block_quote_end_index = (
                    grab_bag.indent_used_by_container + 1 if ind2 else 0
                )
                extra_indent = 1 if grab_bag.container_depth else 0
                list_token = cast(
                    ListStartMarkdownToken,
                    parser_state.token_stack[ind].matching_markdown_token,
                )
                POGGER.debug("__handle_leading_whitespace>>list_token>>$", list_token)
                list_token.add_leading_spaces(
                    position_marker.text_to_parse[
                        block_quote_end_index : grab_bag.indent_already_processed
                        + extra_indent
                    ]
                )
                POGGER.debug("__handle_leading_whitespace>>list_token>>$", list_token)

    @staticmethod
    def __handle_leading_whitespace_loop(
        parser_state: ParserState,
        i: int,
        remaining_whitespace: str,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[bool, int, str]:
        # POGGER.debug(
        #     "$>remaining_whitespace:$:($)",
        #     i,
        #     remaining_whitespace,
        #     len(remaining_whitespace),
        # )
        # POGGER.debug("$>stack:$:", i, parser_state.token_stack[i])
        # POGGER.debug(
        #     "$>token:$:", i, parser_state.token_stack[i].matching_markdown_token
        # )
        inner_token, xx, found_stack_index, remaining_whitespace = (
            ContainerBlockNonLeafProcessor.__handle_leading_whitespace_loop_a(
                parser_state, i, remaining_whitespace
            )
        )
        if xx:
            return True, found_stack_index, remaining_whitespace
        list_token = cast(ListStartMarkdownToken, inner_token)
        assert (
            grab_bag.indent_used_by_container >= 0
        ), "If using a container, should have processed indent."
        remaining_indent = list_token.indent_level - (
            grab_bag.indent_already_processed + grab_bag.indent_used_by_container
        )
        # POGGER.debug("$>remaining_indent:$:", i, remaining_indent)
        left_whitespace = remaining_whitespace[:remaining_indent]
        # POGGER.debug("$>left_whitespace:$:", i, left_whitespace)
        if len(left_whitespace) < remaining_indent:
            if len(remaining_whitespace) >= 4:
                found_stack_index = i
            # POGGER.debug("3-->")
            return True, found_stack_index, remaining_whitespace
        remaining_whitespace = remaining_whitespace[remaining_indent:]
        grab_bag.indent_already_processed = list_token.indent_level
        return False, found_stack_index, remaining_whitespace

    @staticmethod
    def __handle_leading_whitespace_loop_a(
        parser_state: ParserState, i: int, remaining_whitespace: str
    ) -> Tuple[Optional[MarkdownToken], bool, int, str]:
        found_stack_index = 0
        if (inner_token := parser_state.token_stack[i].matching_markdown_token) is None:
            assert (
                parser_state.token_stack[i].was_link_definition_started
                or parser_state.token_stack[i].was_table_block_started
            ), "If there is no matching stack token, this must be a link definition or table."
            return inner_token, True, 0, remaining_whitespace
        if inner_token.is_block_quote_start:
            start_bq_index = remaining_whitespace.find(">")
            if start_bq_index < 0 or start_bq_index >= 4:
                # POGGER.debug("1-->$>start_bq_index:$:", i, start_bq_index)
                # POGGER.debug("$>remaining_whitespace:$:", i, remaining_whitespace)
                if len(remaining_whitespace) >= 4:
                    found_stack_index = i
                return inner_token, True, found_stack_index, remaining_whitespace
            raise AssertionError()
        if not inner_token.is_list_start:
            # POGGER.debug("2-->")
            if len(remaining_whitespace) >= 4:
                found_stack_index = (
                    i + 1 if parser_state.token_stack[i].is_indented_code_block else i
                )
            return inner_token, True, found_stack_index, remaining_whitespace
        return inner_token, False, found_stack_index, remaining_whitespace

    @staticmethod
    def __process_lazy_lines(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> None:
        assert grab_bag.is_leaf_tokens_empty(), "No leaf tokens should be present yet."
        POGGER.debug("clt>>lazy-check")

        after_ws_index, ex_whitespace = ParserHelper.extract_spaces_verified(
            grab_bag.line_to_parse, 0
        )
        remaining_line = grab_bag.line_to_parse[after_ws_index:]
        POGGER.debug("__process_lazy_lines>>mod->ltp>$<", remaining_line)
        POGGER.debug("__process_lazy_lines>>mod->ews>$<", ex_whitespace)

        (
            lazy_tokens,
            grab_bag.block_quote_data,
            grab_bag.was_paragraph_continuation,
        ) = BlockQuoteProcessor.check_for_lazy_handling(
            parser_state,
            position_marker,
            grab_bag.block_quote_data,
            remaining_line,
            ex_whitespace,
            grab_bag.was_paragraph_continuation,
            grab_bag.original_line,
        )
        grab_bag.extend_container_tokens(lazy_tokens)

    @staticmethod
    def __get_block_start_index(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[bool, int, bool]:
        POGGER.debug("text_to_parse>$<", position_marker.text_to_parse)
        POGGER.debug("index_number>$<", position_marker.index_number)
        POGGER.debug("index_indent>$<", position_marker.index_indent)
        new_position_marker = PositionMarker(
            position_marker.line_number,
            grab_bag.start_index,
            position_marker.text_to_parse,
        )
        POGGER.debug("text_to_parse>$<", new_position_marker.text_to_parse)
        POGGER.debug("index_number>$<", new_position_marker.index_number)
        POGGER.debug("container_start_bq_count>$<", grab_bag.container_start_bq_count)
        assert (
            grab_bag.container_start_bq_count is not None
        ), "If here, we should have a count of bq starts."
        (
            did_process,
            block_index,
            block_leaf_tokens,
            block_container_tokens,
            avoid_block_starts,
        ) = BlockQuoteProcessor.handle_block_quote_block(
            parser_state,
            new_position_marker,
            grab_bag,
        )
        grab_bag.extend_container_tokens(block_container_tokens)
        grab_bag.extend_leaf_tokens(block_leaf_tokens)

        if grab_bag.requeue_line_info:
            POGGER.debug(">>requeuing lines after looking for block start. returning.")

        if grab_bag.did_blank and not grab_bag.requeue_line_info:
            ContainerBlockNonLeafProcessor.__get_block_start_index_handle_blank_line(
                parser_state, grab_bag, block_leaf_tokens
            )

        grab_bag.can_continue = (
            not grab_bag.requeue_line_info and not grab_bag.did_blank
        )
        return (
            did_process,
            block_index,
            avoid_block_starts,
        )

    @staticmethod
    def __get_block_start_index_handle_blank_line(
        parser_state: ParserState,
        grab_bag: ContainerGrabBag,
        block_leaf_tokens: List[MarkdownToken],
    ) -> None:
        assert (
            block_leaf_tokens and block_leaf_tokens[-1].is_blank_line
        ), "should be a blank at the end"
        POGGER.debug(">>already handled blank line. returning.")
        grab_bag.extend_container_tokens_with_leaf_tokens()
        stack_index = len(parser_state.token_stack) - 1
        if (
            stack_index > 2
            and parser_state.token_stack[stack_index].is_block_quote
            and parser_state.token_stack[stack_index - 1].is_block_quote
            and parser_state.token_stack[stack_index - 2].is_list
        ):
            list_stack_index = stack_index - 2
            list_token = cast(
                ListStartMarkdownToken,
                parser_state.token_stack[list_stack_index].matching_markdown_token,
            )
            assert list_token is not None

            if list_token.line_number != block_leaf_tokens[-1].line_number:
                ContainerBlockNonLeafProcessor.__sdf(
                    parser_state, grab_bag, list_stack_index, list_token
                )

    @staticmethod
    def __sdf(
        parser_state: ParserState,
        grab_bag: ContainerGrabBag,
        list_stack_index: int,
        list_token: ListStartMarkdownToken,
    ) -> None:
        # do_add_leading_space = True
        # # for i in range(len(parser_state.block_copy)):
        # #     j = parser_state.block_copy[i]
        # #     if j.line_number == list_token.line_number and j.column_number == list_token.column_number:
        # #         break
        # # fg = parser_state.block_copy[i]
        # # fgg = fg.leading_spaces.count("\n") if fg.leading_spaces else 0
        # # list_tokeng = list_token.leading_spaces.count("\n")
        # # ww = (fgg == list_tokeng)
        # if do_add_leading_space:
        bq_count = grab_bag.block_quote_data.current_count
        last_active_block_quote_index = 0
        while bq_count > 0 and last_active_block_quote_index < len(
            parser_state.block_copy
        ):
            next_token = parser_state.block_copy[last_active_block_quote_index]
            assert next_token is not None
            if next_token.is_block_quote_start:
                bq_count -= 1
            last_active_block_quote_index += 1
        do_add_leading_space = list_stack_index < last_active_block_quote_index
        if do_add_leading_space:
            POGGER.debug(
                "__get_block_start_index_handle_blank_line>>list_token>>$",
                list_token,
            )
            list_token.add_leading_spaces("")
            POGGER.debug(
                "__get_block_start_index_handle_blank_line>>list_token>>$",
                list_token,
            )

    @staticmethod
    def __process_list_in_progress(
        parser_state: ParserState,
        index_indent: int,
        grab_bag: ContainerGrabBag,
    ) -> bool:
        did_process, ind = LeafBlockProcessorParagraph.check_for_list_in_process(
            parser_state
        )
        if did_process:
            assert (
                not grab_bag.container_tokens
            ), "Should not have any container tokens yet."
            POGGER.debug("clt>>list-in-progress")
            resultant_tokens = ListBlockProcessor.list_in_process(
                parser_state,
                ind,
                index_indent,
                grab_bag,
            )
            grab_bag.extend_container_tokens(resultant_tokens)

        return did_process

    @staticmethod
    def __get_list_start_index(
        parser_state: ParserState,
        position_marker: PositionMarker,
        is_ulist: bool,
        did_process: bool,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[bool, int]:
        """
        Note: This is one of the more heavily traffic functions in the
        parser.  Debugging should be uncommented only if needed.
        """
        new_position_marker = PositionMarker(
            position_marker.line_number, grab_bag.start_index, grab_bag.line_to_parse
        )

        POGGER.debug(
            "pre-list>>#$#$#$#",
            position_marker.index_number,
            position_marker.index_indent,
            position_marker.text_to_parse,
        )
        POGGER.debug(
            "pre-list>>#$#$#$#",
            new_position_marker.index_number,
            new_position_marker.index_indent,
            new_position_marker.text_to_parse,
        )
        new_list_index = -1
        if not did_process:
            assert (
                grab_bag.adj_ws is not None
            ), "If we have a list, we must have adjusted whitespace."
            assert (
                grab_bag.extracted_whitespace is not None
            ), "Extracted whitespace must be defined at this point."
            assert (
                grab_bag.removed_chars_at_start_of_line is not None
            ), "If we have a list, we must have removed characters."
            (
                did_process,
                new_list_index,
                new_line_to_parse,
                resultant_tokens,
            ) = ListBlockProcessor.handle_list_block(
                is_ulist,
                parser_state,
                new_position_marker,
                grab_bag,
            )
            # POGGER.debug_with_visible_whitespace("handle_list_block>$", resultant_tokens)
            if not grab_bag.requeue_line_info:
                assert (
                    new_line_to_parse is not None
                ), "If not requeuing, must have more line to parse."
                grab_bag.line_to_parse = new_line_to_parse
                grab_bag.extend_container_tokens(resultant_tokens)
        POGGER.debug(
            "post-ulist>>#$#$#$#",
            position_marker.index_number,
            position_marker.index_indent,
            position_marker.text_to_parse,
        )
        POGGER.debug(
            "post-ulist>>#$#$#$#",
            new_position_marker.index_number,
            new_position_marker.index_indent,
            new_position_marker.text_to_parse,
        )

        if grab_bag.requeue_line_info:
            POGGER.debug(
                ">>requeuing lines after looking for ordered list start. returning."
            )

        grab_bag.can_continue = not grab_bag.requeue_line_info
        return (
            did_process,
            new_list_index,
        )



"""
Module to help with the parsing of inline elements at the end of the line.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.inline.inline_backslash_helper import InlineBackslashHelper
from pymarkdown.inline.inline_request import InlineRequest
from pymarkdown.inline.inline_response import InlineResponse
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.hard_break_markdown_token import HardBreakMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods
class InlineLineEndHelper:
    """
    Class to help with the parsing of inline elements at the end of the line.
    """

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_line_end(
        remaining_line: str,
        tabified_remaining_line: Optional[str],
        end_string: Optional[str],
        current_string: str,
        inline_blocks: List[MarkdownToken],
        is_setext: bool,
        line_number: int,
        column_number: int,
        coalesced_stack: List[MarkdownToken],
        tabified_text: Optional[str],
        inline_request: InlineRequest,
    ) -> Tuple[
        str, Optional[str], List[MarkdownToken], str, Optional[str], str, Optional[str]
    ]:
        """
        Handle the inline case of having the end of line character encountered.
        """
        new_tokens: List[MarkdownToken] = []

        # POGGER.debug(">>current_string>>$>>", current_string)
        # POGGER.debug(">>end_string>>$>>", end_string)
        # POGGER.debug(">>remaining_line>>$>>", remaining_line)
        # POGGER.debug(">>tabified_remaining_line>>$>>", tabified_remaining_line)
        (
            removed_end_whitespace,
            remaining_line,
        ) = InlineLineEndHelper.__setup_for_select_line_ending(
            tabified_remaining_line, is_setext, remaining_line
        )
        # POGGER.debug(">>line_to_use>>$>>", line_to_use)
        # POGGER.debug(">>current_string>>$>>", current_string)

        (
            current_string,
            whitespace_to_add,
            append_to_current_string,
            end_string,
            remaining_line,
            tabified_remaining_line,
        ) = InlineLineEndHelper.__select_line_ending(
            new_tokens,
            line_number,
            column_number + len(remaining_line),
            current_string,
            removed_end_whitespace,
            end_string,
            remaining_line,
            inline_blocks,
            is_setext,
            tabified_text,
            inline_request,
            tabified_remaining_line,
        )

        InlineLineEndHelper.__handle_line_end_adjust_block_quote(coalesced_stack)

        return (
            append_to_current_string,
            whitespace_to_add,
            new_tokens,
            remaining_line,
            end_string,
            current_string,
            tabified_remaining_line,
        )

    # pylint: enable=too-many-arguments
    @staticmethod
    def __setup_for_select_line_ending(
        tabified_remaining_line: Optional[str], is_setext: bool, remaining_line: str
    ) -> Tuple[str, str]:
        line_to_use = (
            tabified_remaining_line
            if tabified_remaining_line and is_setext
            else remaining_line
        )
        _, last_non_whitespace_index = ParserHelper.collect_backwards_while_character(
            line_to_use, -1, " "
        )
        removed_end_whitespace = line_to_use[last_non_whitespace_index:]
        remaining_line = line_to_use[:last_non_whitespace_index]
        return removed_end_whitespace, remaining_line

    @staticmethod
    def __handle_line_end_adjust_block_quote(
        coalesced_stack: List[MarkdownToken],
    ) -> None:
        if coalesced_stack and coalesced_stack[-1].is_block_quote_start:
            block_quote_token = cast(BlockQuoteMarkdownToken, coalesced_stack[-1])
            block_quote_token.leading_text_index += 1

    @staticmethod
    def __is_proper_hard_break(
        current_string: str, removed_end_whitespace_size: int
    ) -> bool:
        POGGER.debug("__is_proper_hard_break>>current_string>>$>>", current_string)
        POGGER.debug("removed_end_whitespace_size>>$>>", removed_end_whitespace_size)

        current_string_size = len(current_string)
        if (
            removed_end_whitespace_size == 0
            and current_string_size
            and current_string[current_string_size - 1]
            == InlineBackslashHelper.backslash_character
        ):
            POGGER.debug(">>$<<", current_string)
            modified_current_string = current_string[:-1]
            is_proper_hard_break = modified_current_string[-2:] != "\\\b"
            POGGER.debug(">>$<<", is_proper_hard_break)
        else:
            is_proper_hard_break = False

        POGGER.debug("__is_proper_hard_break>>$>>", is_proper_hard_break)
        return is_proper_hard_break

    @staticmethod
    def __select_line_end_hard_break(
        new_tokens: List[MarkdownToken],
        line_number: int,
        adj_hard_column: int,
        current_string: str,
    ) -> Tuple[str, Optional[str], str]:
        POGGER.debug(">>proper hard break")
        new_tokens.append(
            HardBreakMarkdownToken(
                InlineBackslashHelper.backslash_character,
                line_number,
                adj_hard_column - 1,
            )
        )
        current_string, whitespace_to_add = current_string[:-1], None
        append_to_current_string = ""
        return current_string, whitespace_to_add, append_to_current_string

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __select_line_ending(
        new_tokens: List[MarkdownToken],
        line_number: int,
        adj_hard_column: int,
        current_string: str,
        removed_end_whitespace: str,
        end_string: Optional[str],
        remaining_line: str,
        inline_blocks: List[MarkdownToken],
        is_setext: bool,
        tabified_text: Optional[str],
        inline_request: InlineRequest,
        tabified_remaining_line: Optional[str],
    ) -> Tuple[str, Optional[str], str, Optional[str], str, Optional[str]]:
        # POGGER.debug(">>removed_end_whitespace>:$:<", removed_end_whitespace)
        # POGGER.debug(">>tabified_text>:$:<", tabified_text)
        # POGGER.debug(
        #     ">>inline_request.tabified_text>:$:<", inline_request.tabified_text
        # )
        # POGGER.debug(
        #     ">>inline_request.tabified_remaining_line>:$:<",
        #     inline_request.tabified_remaining_line,
        # )
        append_to_current_string = ParserHelper.newline_character
        whitespace_to_add: Optional[str] = None

        removed_end_whitespace_size = len(removed_end_whitespace)
        # POGGER.debug(
        #     ">>len(r_e_w)>>$>>rem>>$>>",
        #     removed_end_whitespace_size,
        #     remaining_line,
        # )

        is_proper_end = not tabified_text or (
            tabified_remaining_line and tabified_remaining_line.endswith("  ")
        )

        if InlineLineEndHelper.__is_proper_hard_break(
            current_string, removed_end_whitespace_size
        ):
            current_string, whitespace_to_add, append_to_current_string = (
                InlineLineEndHelper.__select_line_end_hard_break(
                    new_tokens, line_number, adj_hard_column, current_string
                )
            )
        elif removed_end_whitespace_size >= 2 and is_proper_end:
            POGGER.debug(">>whitespace hard break")
            new_tokens.append(
                HardBreakMarkdownToken(
                    removed_end_whitespace, line_number, adj_hard_column
                )
            )
            whitespace_to_add = None
            append_to_current_string = ""

            if tabified_remaining_line is not None:
                number_collected_characters, start_index = (
                    ParserHelper.collect_backwards_while_character_verified(
                        tabified_remaining_line, len(tabified_remaining_line), " "
                    )
                )
                assert (
                    number_collected_characters >= 2
                ), "Based on the elif statement, should collect at least 2 characters."
                tabified_remaining_line = tabified_remaining_line[:start_index]
        else:
            POGGER.debug(">>normal end")
            # POGGER.debug("current_string>:$:<", current_string)
            # POGGER.debug("removed_end_whitespace>:$:<", removed_end_whitespace)
            # POGGER.debug("end_string>:$:<", end_string)
            # POGGER.debug("remaining_line>:$:<", remaining_line)
            (
                end_string,
                remaining_line,
            ) = InlineLineEndHelper.__select_line_ending_normal(
                is_setext,
                inline_blocks,
                current_string,
                removed_end_whitespace,
                inline_request.tabified_remaining_line,
                end_string,
                remaining_line,
            )

        # POGGER.debug(
        #     "<<append_to_current_string<<$<<",
        #     append_to_current_string,
        # )
        # POGGER.debug(
        #     "<<whitespace_to_add<<$<<",
        #     whitespace_to_add,
        # )
        # POGGER.debug("<<remaining_line<<$<<", remaining_line)
        # POGGER.debug("<<end_string<<$<<", end_string)
        # POGGER.debug("<<current_string<<$<<", current_string)
        return (
            current_string,
            whitespace_to_add,
            append_to_current_string,
            end_string,
            remaining_line,
            tabified_remaining_line,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __select_line_ending_normal(
        is_setext: bool,
        inline_blocks: List[MarkdownToken],
        current_string: str,
        removed_end_whitespace: str,
        tabified_remaining_line: Optional[str],
        end_string: Optional[str],
        remaining_line: str,
    ) -> Tuple[str, str]:
        # POGGER.debug("<<is_setext<<$<<", is_setext)
        # POGGER.debug("<<inline_blocks<<$<<", inline_blocks)
        # POGGER.debug("<<current_string<<$<<", current_string)
        POGGER.debug("<<remaining_line<<$<<", remaining_line)
        POGGER.debug("<<end_string<<$<<", end_string)
        POGGER.debug("<<removed_end_whitespace<<$<<", removed_end_whitespace)
        if (
            is_setext
            and inline_blocks
            and inline_blocks[-1].is_inline_hard_break
            and not current_string
        ):
            new_index, ex_ws = ParserHelper.extract_spaces(remaining_line, 0)
            # POGGER.debug("<<new_index<<$<<", new_index)
            # POGGER.debug("<<ex_ws<<$<<", ex_ws)
            end_string = (
                f"{ex_ws}{ParserHelper.whitespace_split_character}" if new_index else ""
            )
            remaining_line = remaining_line[new_index:]
        if not is_setext and tabified_remaining_line and removed_end_whitespace:
            POGGER.debug("<<tabified_remaining_line>:$:<", tabified_remaining_line)
            POGGER.debug("<<removed_end_whitespace>:$:<", removed_end_whitespace)
            removed_end_whitespace, _ = TabHelper.find_detabify_string_ex(
                tabified_remaining_line, removed_end_whitespace
            )
            POGGER.debug("<<removed_end_whitespace<<$<<", removed_end_whitespace)

        POGGER.debug("<<end_string<<$<<", end_string)
        end_string = (
            f"{removed_end_whitespace}{ParserHelper.newline_character}"
            if end_string is None
            else f"{end_string}{removed_end_whitespace}{ParserHelper.newline_character}"
        )
        POGGER.debug("<<end_string<<$<<", end_string)
        return end_string, remaining_line

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def process_inline_new_line(
        source_text: str,
        next_index: int,
        inline_response: InlineResponse,
        remaining_line: str,
        end_string: Optional[str],
        current_string: str,
        inline_blocks: List[MarkdownToken],
        is_setext: bool,
        line_number: int,
        column_number: int,
        coalesced_stack: List[MarkdownToken],
        whitespace_to_recombine: Optional[str],
        para_owner: Optional[ParagraphMarkdownToken],
        tabified_text: Optional[str],
        inline_request: InlineRequest,
        tabified_remaining_line: Optional[str],
    ) -> Tuple[Optional[str], str, Optional[str], str, bool, Optional[str]]:
        """
        Process a new line character.
        """
        assert (
            source_text[next_index] == ParserHelper.newline_character
        ), "Processing a newline, so the indexed character must be one."
        # POGGER.debug("end_string>:$:<", end_string)
        # POGGER.debug("remaining_line>:$:<", remaining_line)
        # POGGER.debug("tabified_remaining_line>:$:<", tabified_remaining_line)
        # POGGER.debug("tabified_text>:$:<", tabified_text)
        # POGGER.debug(
        #     "inline_request.tabified_remaining_line>:$:<",
        #     inline_request.tabified_remaining_line,
        # )
        (
            inline_response.new_string,
            whitespace_to_add,
            inline_response.new_tokens,
            remaining_line,
            end_string,
            current_string,
            tabified_remaining_line,
        ) = InlineLineEndHelper.__handle_line_end(
            remaining_line,
            tabified_remaining_line,
            end_string,
            current_string,
            inline_blocks,
            is_setext,
            line_number,
            column_number,
            coalesced_stack,
            tabified_text,
            inline_request,
        )
        inline_response.new_index = next_index + 1
        # POGGER.debug("end_string>:$:<", end_string)
        # POGGER.debug("remaining_line>:$:<", remaining_line)
        # POGGER.debug(
        #     "handle_line_end>>new_tokens>>$<<",
        #     inline_response.new_tokens,
        # )

        if not inline_response.new_tokens:
            # POGGER.debug("ws")
            assert (
                end_string is not None
            ), "If no new tokens, an end_string must be produced."
            end_string = InlineLineEndHelper.__add_recombined_whitespace(
                bool(whitespace_to_recombine),
                source_text,
                inline_response,
                end_string,
                is_setext,
            )
            # POGGER.debug(
            #     "3<<end_string<<$<<",
            #     end_string,
            # )
            # POGGER.debug("ws>$<", end_string)
        # POGGER.debug(
        #     "handle_line_end>>$<<", source_text[inline_response.new_index :]
        # )
        # POGGER.debug(
        #     "end_string(after)>>$<<",
        #     end_string,
        # )
        # POGGER.debug(">>line_number>>$<<", line_number)
        # POGGER.debug(">>column_number>>$<<", column_number)
        if para_owner:
            # POGGER.debug(">>para_owner.rehydrate_index>>$<<", para_owner.rehydrate_index)
            para_owner.rehydrate_index += 1
            # POGGER.debug(">>para_owner.rehydrate_index>>$<<", para_owner.rehydrate_index)

        if tabified_remaining_line and end_string is not None and len(end_string) > 1:
            # assert end_string is not None
            tabified_remaining_line = InlineLineEndHelper.__clean_up_new_line(
                end_string, is_setext, tabified_remaining_line
            )

        return (
            whitespace_to_add,
            remaining_line,
            end_string,
            current_string,
            True,
            tabified_remaining_line,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __clean_up_new_line(
        end_string: str, is_setext: bool, tabified_remaining_line: str
    ) -> str:
        POGGER.debug("end_string>$<", end_string)
        POGGER.debug("tabified_remaining_line>$<", tabified_remaining_line)
        assert end_string[-1] in [
            "\n",
            ParserHelper.whitespace_split_character,
        ], "Due to SetExt paragraphs, can be either."
        if end_string[-1] == ParserHelper.whitespace_split_character:
            newline_index = end_string.rfind("\n")
            assert (
                newline_index != -1
            ), "if we hit a split character, must find the newline"
            end_suffix = end_string[:newline_index]
        else:
            end_suffix = end_string[:-1]
        POGGER.debug("end_suffix>$<", end_suffix)
        newline_index = end_suffix.rfind("\n")
        if is_setext:
            special_index = end_suffix.rfind(ParserHelper.whitespace_split_character)
            if special_index != -1 or newline_index != -1:
                max_index = max(special_index, newline_index)
                end_suffix = end_suffix[max_index + 1 :]
            else:
                assert special_index == -1, "other side of the if statement"
                assert newline_index != 1, "other side of the if statement"
                end_suffix = end_suffix[newline_index + 1 :]
        elif newline_index != -1:
            end_suffix = end_suffix[newline_index + 1 :]

        POGGER.debug("tabified_remaining_line>$<", tabified_remaining_line)
        POGGER.debug("end_suffix>$<", end_suffix)
        assert tabified_remaining_line.endswith(
            end_suffix
        ), "Tabified line should end with the newly computed suffix."
        if end_suffix:
            tabified_remaining_line = tabified_remaining_line[: -(len(end_suffix))]
        POGGER.debug("tabified_remaining_line>$<", tabified_remaining_line)
        return tabified_remaining_line

    @staticmethod
    def __add_recombined_whitespace(
        did_recombine: bool,
        source_text: str,
        inline_response: InlineResponse,
        end_string: str,
        is_setext: bool,
    ) -> str:
        POGGER.debug("__arw>>did_recombine>>$>>", did_recombine)
        POGGER.debug(
            "__arw>>end_string>>$>>",
            end_string,
        )
        if did_recombine:
            POGGER.debug(
                "__arw>>source_text>>$>>",
                source_text,
            )
            assert (
                inline_response.new_index is not None
            ), "new_index should be defined by this point."
            new_index, extracted_whitespace = ParserHelper.extract_spaces_verified(
                source_text, inline_response.new_index
            )
            POGGER.debug("__arw>>$>>", source_text[: inline_response.new_index])
            POGGER.debug("__arw>>$>>", source_text[inline_response.new_index :])
            POGGER.debug(
                "__arw>>extracted_whitespace>>$>>",
                extracted_whitespace,
            )
            assert (
                is_setext
            ), "If trying to recombine and whitespace is found, must be text in an SetExt heading."
            if extracted_whitespace:
                inline_response.new_index = new_index
                end_string = f"{end_string}{extracted_whitespace}{ParserHelper.whitespace_split_character}"
                POGGER.debug(
                    "__arw>>end_string>>$>>",
                    end_string,
                )
        return end_string


# pylint: enable=too-few-public-methods



"""
Module to provide classes to deal with plugins.
"""

import argparse
import inspect
import logging
import os
import re
import sys
from io import TextIOWrapper
from typing import Any, Dict, List, Optional, Set, Tuple

from application_properties import ApplicationProperties, ApplicationPropertiesFacade
from columnar import columnar

from pymarkdown.extensions.pragma_token import (
    GeneralPragmaDisableStart,
    PragmaExtension,
)
from pymarkdown.general.main_presentation import MainPresentation
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.bad_plugin_error import BadPluginError
from pymarkdown.plugin_manager.fix_line_record import FixLineRecord
from pymarkdown.plugin_manager.fix_token_record import FixTokenRecord
from pymarkdown.plugin_manager.found_plugin import FoundPlugin
from pymarkdown.plugin_manager.plugin_details import PluginDetailsV2, QueryConfigItem
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.plugin_scan_failure import PluginScanFailure
from pymarkdown.plugin_manager.replace_tokens_record import ReplaceTokensRecord
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.return_code_helper import ApplicationResult
from pymarkdown.tokens.markdown_token import MarkdownToken

LOGGER = logging.getLogger(__name__)

# pylint: disable=too-many-lines


# pylint: disable=too-many-instance-attributes
class PluginManager:
    """
    Manager object to take care of load and accessing plugin modules.
    """

    __plugin_prefix = "plugins"
    __root_subparser_name = "pm_subcommand"
    __argparse_subparser: Optional[argparse.ArgumentParser] = None
    __id_regex = re.compile("^[a-z]{2,3}\\d{3}$")
    __name_regex = re.compile("^[a-zA-Z0-9][a-zA-Z0-9-]*[a-zA-Z0-9]$")
    __filter_regex = re.compile("^[a-zA-Z0-9-]+$")
    __version_regex = re.compile("^(0|[1-9]\\d*)\\.(0|[1-9]\\d*)\\.(0|[1-9]\\d*)$")
    __disable_rules_wildcard = "*"

    def __init__(
        self,
        presentation: MainPresentation,
    ) -> None:
        (
            self.number_of_scan_failures,
            self.number_of_pragma_failures,
            self.__show_stack_trace,
            self.__show_fix_debug,
        ) = (0, 0, False, False)
        self.__loaded_classes: List[Tuple[RulePlugin, str]] = []

        self.__presentation = presentation

        self.__document_pragmas: Dict[int, Set[str]] = {}
        self.__document_pragma_ranges: List[Tuple[int, int, Set[str]]]
        self.__general_pragma_ranges: List[Tuple[int, int, str]]
        self.__pragma_line_numbers: List[int]

        self.__registered_plugins: List[FoundPlugin] = []
        self.__enabled_plugins: List[FoundPlugin] = []
        self.__enabled_plugins_for_starting_new_file: List[FoundPlugin] = []
        self.__enabled_plugins_for_next_token: List[FoundPlugin] = []
        self.__enabled_plugins_for_next_line: List[FoundPlugin] = []
        self.__enabled_plugins_for_completed_file: List[FoundPlugin] = []
        self.__all_ids: Dict[str, FoundPlugin] = {}
        self.__properties: Optional[ApplicationProperties] = None

    # pylint: disable=too-many-arguments
    def initialize(
        self,
        directory_to_search: str,
        additional_paths: List[str],
        enable_rules_from_command_line: str,
        disable_rules_from_command_line: str,
        properties: ApplicationProperties,
        show_stack_trace: bool,
        show_fix_debug: bool,
    ) -> None:
        """
        Initializes the manager by scanning for plugins, loading them, and registering them.
        """
        (
            self.number_of_scan_failures,
            self.number_of_pragma_failures,
            self.__loaded_classes,
            self.__show_stack_trace,
            self.__show_fix_debug,
            self.__properties,
        ) = (0, 0, [], show_stack_trace, show_fix_debug, properties)

        plugin_files = self.__find_eligible_plugins_in_directory(directory_to_search)
        self.__load_plugins(directory_to_search, plugin_files)

        all_additional_paths = list(additional_paths or [])
        if more_paths := properties.get_string_property("plugins.additional_paths"):
            all_additional_paths.extend(more_paths.split(","))

        if all_additional_paths:
            for next_additional_plugin in all_additional_paths:
                if not os.path.exists(next_additional_plugin):
                    formatted_message = (
                        f"Plugin path '{next_additional_plugin}' does not exist."
                    )
                    raise BadPluginError(formatted_message=formatted_message)
                if os.path.isdir(next_additional_plugin):
                    plugin_files = self.__find_eligible_plugins_in_directory(
                        next_additional_plugin
                    )
                    self.__load_plugins(next_additional_plugin, plugin_files)
                else:
                    self.__load_plugins(
                        os.path.dirname(next_additional_plugin),
                        [os.path.basename(next_additional_plugin)],
                    )

        self.__register_plugins(
            enable_rules_from_command_line, disable_rules_from_command_line, properties
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def argparse_subparser_name() -> str:
        """
        Gets the name of the subparser used to handle these plugins.
        """
        return "plugins"

    @staticmethod
    def __list_filter_type(argument: str) -> str:
        test_argument = argument.replace("*", "").replace("?", "")
        if PluginManager.__filter_regex.match(test_argument):
            return argument
        raise argparse.ArgumentTypeError(
            f"Value '{argument}' is not a valid pattern for an id or a name."
        )

    @staticmethod
    def __info_filter_type(argument: str) -> str:
        if PluginManager.__id_regex.match(argument) or PluginManager.__name_regex.match(
            argument
        ):
            return argument
        raise argparse.ArgumentTypeError(
            f"Value '{argument}' is not a valid id or name."
        )

    @staticmethod
    def add_argparse_subparser(subparsers: argparse._SubParsersAction) -> None:  # type: ignore
        """
        Populate the argparse tree to allow for plugin support.
        """

        new_sub_parser = subparsers.add_parser(
            PluginManager.argparse_subparser_name(), help="plugin commands"
        )
        PluginManager.__argparse_subparser = new_sub_parser
        plugin_subparsers = new_sub_parser.add_subparsers(
            dest=PluginManager.__root_subparser_name
        )

        sub_sub_parser = plugin_subparsers.add_parser(
            "list", help="list the available plugins"
        )
        sub_sub_parser.add_argument(
            "--all",
            dest="show_all",
            action="store_true",
            default=False,
            help="show all loaded plugins (default is False)",
        )
        sub_sub_parser.add_argument(
            dest="list_filter",
            default=None,
            help="filter",
            nargs="?",
            type=PluginManager.__list_filter_type,
        )
        sub_sub_parser = plugin_subparsers.add_parser(
            "info", help="information on a specific plugin"
        )
        sub_sub_parser.add_argument(
            dest="info_filter",
            default=None,
            type=PluginManager.__info_filter_type,
            help="an id",
        )

    def __add_row_for_next_plugin(
        self,
        next_plugin_id: str,
        args: argparse.Namespace,
        show_rows: List[List[str]],
        list_re: Optional[re.Pattern[str]],
    ) -> None:
        # if next_plugin_id.startswith("md9"):
        #     continue
        next_plugin_list: List[FoundPlugin] = []
        for next_plugin in self.__registered_plugins:
            if next_plugin.plugin_id == next_plugin_id:
                next_plugin_list.append(next_plugin)
        assert len(next_plugin_list) == 1
        next_plugin = next_plugin_list[0]
        if next_plugin.plugin_version != "0.0.0" or args.show_all:
            self.__show_row_if_matches(list_re, next_plugin_id, next_plugin, show_rows)

    def __handle_argparse_subparser_list(
        self, args: argparse.Namespace
    ) -> ApplicationResult:
        list_re = (
            re.compile(
                "^" + args.list_filter.replace("*", ".*").replace("?", ".") + "$"
            )
            if args.list_filter
            else None
        )

        show_rows: List[List[str]] = []
        ids = self.all_plugin_ids
        ids.sort()
        for next_plugin_id in ids:
            self.__add_row_for_next_plugin(next_plugin_id, args, show_rows, list_re)
        if show_rows:
            headers = [
                "id",
                "names",
                "enabled\n(default)",
                "enabled\n(current)",
                "version",
                "fix",
            ]
            self.__print_columnar_data(headers, show_rows)
            return ApplicationResult.SUCCESS
        self.__presentation.print_system_error(
            f"No plugin rule identifiers matches the pattern '{args.list_filter}'."
        )
        return ApplicationResult.NO_FILES_TO_SCAN

    def __show_row_if_matches(
        self,
        list_re: Optional[re.Pattern[str]],
        next_plugin_id: str,
        next_plugin: FoundPlugin,
        show_rows: List[List[str]],
    ) -> None:
        does_match = True
        if list_re:
            does_match = list_re.match(next_plugin_id) is not None
            if not does_match:
                for next_name in next_plugin.plugin_names:
                    does_match = list_re.match(next_name) is not None
                    if does_match:
                        break
        if does_match:
            is_enabled_now = next_plugin in self.__enabled_plugins
            display_row = [
                next_plugin_id,
                ", ".join(next_plugin.plugin_names),
                str(next_plugin.plugin_enabled_by_default),
                str(is_enabled_now),
                next_plugin.plugin_version,
                "Yes" if next_plugin.plugin_supports_fix else "No",
            ]
            show_rows.append(display_row)

    def __print_columnar_data(
        self, headers: List[str], show_rows: List[List[str]]
    ) -> None:
        table = columnar(show_rows, headers, no_borders=True)
        split_rows = table.split(ParserHelper.newline_character)
        new_rows = [next_row.rstrip() for next_row in split_rows]
        self.__presentation.print_system_output(
            ParserHelper.newline_character.join(new_rows)
        )

    def __handle_argparse_subparser_info(
        self, args: argparse.Namespace
    ) -> ApplicationResult:
        matching_plugins: List[FoundPlugin] = list(
            filter(
                lambda x: args.info_filter in x.plugin_identifiers,
                self.__registered_plugins,
            )
        )
        if not matching_plugins:
            self.__presentation.print_system_error(
                f"Unable to find a plugin with an id or name of '{args.info_filter}'."
            )
            return ApplicationResult.NO_FILES_TO_SCAN

        assert self.__properties is not None, "Should have been set during initialize."
        self.__apply_configuration(matching_plugins[0], self.__properties)

        found_plugin = matching_plugins[0]
        show_rows = [
            ["Id", found_plugin.plugin_id],
            ["Name(s)", ",".join(found_plugin.plugin_names)],
            ["Short Description", found_plugin.plugin_description],
        ]
        if found_plugin.plugin_url:
            next_row = ["Description Url", found_plugin.plugin_url]
            show_rows.append(next_row)

        actual_item_list = []
        if (
            found_plugin.plugin_interface_version == 3
            and found_plugin.plugin_instance.is_query_config_implemented_in_plugin
        ):
            try:
                actual_item_list = found_plugin.plugin_instance.query_config()
            except Exception as this_exception:
                raise BadPluginError(
                    found_plugin.plugin_id,
                    inspect.stack()[0].function,
                    cause=this_exception,
                ) from this_exception

        self.__handle_argparse_subparser_info_display(
            found_plugin, actual_item_list, show_rows
        )
        return ApplicationResult.SUCCESS

    def __handle_argparse_subparser_info_display(
        self,
        found_plugin: FoundPlugin,
        actual_item_list: List[QueryConfigItem],
        show_rows: List[List[str]],
    ) -> None:
        if not actual_item_list and found_plugin.plugin_configuration:
            next_row = ["Configuration Items", found_plugin.plugin_configuration]
            show_rows.append(next_row)

        headers = ["Item", "Description"]
        self.__print_columnar_data(headers, show_rows)

        if actual_item_list:
            show_rows = []
            for next_item in actual_item_list:
                if isinstance(next_item.value, str):
                    value_type = "string"
                    display_value = f'"{next_item.value}"'
                elif isinstance(next_item.value, bool):
                    value_type = "boolean"
                    display_value = str(next_item.value)
                else:
                    value_type = "integer"
                    display_value = str(next_item.value)
                next_row = [next_item.name, value_type, display_value]
                show_rows.append(next_row)

            headers = ["Configuration Item", "Type", "Value"]
            self.__print_columnar_data(headers, show_rows)

    def handle_argparse_subparser(self, args: argparse.Namespace) -> ApplicationResult:
        """
        Handle the parsing for this subparser.
        """
        subparser_value = getattr(args, PluginManager.__root_subparser_name)
        if subparser_value == "list":
            return self.__handle_argparse_subparser_list(args)
        if subparser_value == "info":
            return self.__handle_argparse_subparser_info(args)
        assert PluginManager.__argparse_subparser
        PluginManager.__argparse_subparser.print_help()
        return ApplicationResult.COMMAND_LINE_ERROR

    def log_scan_failure(self, scan_failure: PluginScanFailure) -> None:
        """
        Log the scan failure in the appropriate format.
        """

        rule_id = scan_failure.rule_id.lower()
        if self.__document_pragmas:
            if scan_failure.line_number in self.__document_pragmas:
                id_set = self.__document_pragmas[scan_failure.line_number]
                if rule_id in id_set:
                    return
            if (
                scan_failure.is_error_token_prefaced_by_blank_line
                and (scan_failure.line_number - 1) in self.__document_pragmas
            ):
                id_set = self.__document_pragmas[scan_failure.line_number - 1]
                if rule_id in id_set:
                    return

        if self.__document_pragma_ranges:
            for i, j, k in self.__document_pragma_ranges:
                if i <= scan_failure.line_number <= j and rule_id in k:
                    return

        if self.__general_pragma_ranges:
            for i, j, m in self.__general_pragma_ranges:
                if i <= scan_failure.line_number <= j and rule_id == m:
                    return

        extra_info = (
            f" [{scan_failure.extra_error_information}]"
            if scan_failure.extra_error_information
            else ""
        )
        rule_id = scan_failure.rule_id.upper()

        adjusted_failure = PluginScanFailure(
            scan_failure.scan_file,
            scan_failure.line_number,
            scan_failure.column_number,
            rule_id,
            scan_failure.rule_name,
            scan_failure.rule_description,
            extra_info,
            False,
        )
        self.__presentation.print_scan_failure(adjusted_failure)
        self.number_of_scan_failures += 1

    def log_pragma_failure(
        self, scan_file: str, line_number: int, pragma_error: str
    ) -> None:
        """
        Log the pragma failure in the appropriate format.
        """
        self.__presentation.print_pragma_failure(scan_file, line_number, pragma_error)
        self.number_of_pragma_failures += 1

    def compile_pragmas(self, scan_file: str, pragma_lines: Dict[int, str]) -> None:
        """
        Go through the list of extracted pragmas and compile them.
        """

        active_general_pragmas: Dict[str, GeneralPragmaDisableStart] = {}

        self.__pragma_line_numbers = []
        for next_line_number in pragma_lines:
            PragmaExtension.compile_single_pragma(
                scan_file,
                next_line_number,
                pragma_lines,
                self.__all_ids,
                self.__document_pragmas,
                self.__document_pragma_ranges,
                self.__general_pragma_ranges,
                active_general_pragmas,
                self.log_pragma_failure,
                self.__pragma_line_numbers,
            )
        PragmaExtension.end(active_general_pragmas, self.__general_pragma_ranges)

    @property
    def enabled_plugins(self) -> List[FoundPlugin]:
        """
        Get a list of the plugins that are currently enabled.
        """
        return self.__enabled_plugins

    @classmethod
    def __find_eligible_plugins_in_directory(
        cls, directory_to_search: str
    ) -> List[str]:
        """
        Given a directory to search, scan for eligible modules to load later.
        """

        return [
            x
            for x in os.listdir(directory_to_search)
            if x.endswith(".py") and x[:-3] != "__init__"
        ]

    @classmethod
    def __snake_to_camel(cls, word: str) -> str:
        return "".join(x.capitalize() or "_" for x in word.split("_"))

    def __attempt_to_load_plugin(
        self, next_plugin_module: str, plugin_class_name: str, next_plugin_file: str
    ) -> None:
        """
        Attempt to cleanly load the specified plugin.
        """
        try:
            mod = __import__(next_plugin_module)
        except Exception as this_exception:
            raise BadPluginError(file_name=next_plugin_file) from this_exception

        if not hasattr(mod, plugin_class_name):
            raise BadPluginError(
                file_name=next_plugin_file, class_name=plugin_class_name
            ) from None
        my_class = getattr(mod, plugin_class_name)

        try:
            plugin_class_instance = my_class()
        except Exception as this_exception:
            raise BadPluginError(
                file_name=next_plugin_file,
                class_name=plugin_class_name,
                is_constructor=True,
            ) from this_exception
        self.__loaded_classes.append((plugin_class_instance, next_plugin_file))

    def __load_plugins(self, directory_to_search: str, plugin_files: List[str]) -> None:
        """
        Given an array of discovered modules, load them into the global namespace.
        """

        if os.path.abspath(directory_to_search) not in sys.path:
            sys.path.insert(0, os.path.abspath(directory_to_search))

        for next_plugin_file in plugin_files:
            next_plugin_module = next_plugin_file[:-3]
            plugin_class_name = self.__snake_to_camel(next_plugin_module)
            self.__attempt_to_load_plugin(
                next_plugin_module, plugin_class_name, next_plugin_file
            )

    def __determine_if_plugin_enabled(
        self,
        plugin_object: FoundPlugin,
        command_line_enabled_rules: Set[str],
        command_line_disabled_rules: Set[str],
        properties: ApplicationProperties,
    ) -> bool:
        """
        Given the enable and disable rule values, evaluate the enabled or disabled
        state of the plugin in proper order.
        """

        LOGGER.debug(
            "Plugin '%s', identifiers: %s",
            plugin_object.plugin_id,
            plugin_object.plugin_identifiers,
        )

        new_value = self.__handle_command_line_settings(
            plugin_object, command_line_disabled_rules, command_line_enabled_rules
        )

        default_value = plugin_object.plugin_enabled_by_default
        if new_value is None:
            disable_all_plugins_value = properties.get_boolean_property(
                f"{PluginManager.__plugin_prefix}{properties.separator}selectively_enable_rules",
                default_value=None,
            )
            if disable_all_plugins_value is not None and disable_all_plugins_value:
                default_value = False
                LOGGER.debug(
                    "Plugin '%s' is disabled by the global setting.",
                    plugin_object.plugin_id,
                )
        if new_value is None:
            if plugin_specific_facade := self.__find_configuration_for_plugin(
                plugin_object, properties
            ):
                LOGGER.debug(
                    "Plugins specific configuration found, searching for key 'enabled'."
                )
                new_value = plugin_specific_facade.get_boolean_property(
                    "enabled", default_value=None
                )
                if new_value:
                    LOGGER.debug(
                        "Plugin specific key 'enabled' found, value is '%s'.",
                        str(new_value),
                    )
        if new_value is None:
            LOGGER.debug(
                "No other enable state found, setting to default of '%s'.",
                str(default_value),
            )

        return default_value if new_value is None else new_value

    @classmethod
    def __handle_command_line_settings(
        cls,
        plugin_object: FoundPlugin,
        command_line_disabled_rules: Set[str],
        command_line_enabled_rules: Set[str],
    ) -> Optional[bool]:
        new_value = None
        was_wildcarded = False
        if command_line_disabled_rules:
            LOGGER.debug(
                "Disabled on command line: %s", str(command_line_disabled_rules)
            )
            if PluginManager.__disable_rules_wildcard in command_line_disabled_rules:
                new_value = False
                was_wildcarded = True
            else:
                for next_identifier in plugin_object.plugin_identifiers:
                    if next_identifier in command_line_disabled_rules:
                        new_value = False
                        LOGGER.debug("Plugin is disabled from command line.")
                        break
        if (new_value is None or was_wildcarded) and command_line_enabled_rules:
            LOGGER.debug("Enabled on command line: %s", str(command_line_enabled_rules))
            for next_identifier in plugin_object.plugin_identifiers:
                if next_identifier in command_line_enabled_rules:
                    new_value = True
                    LOGGER.debug("Plugin is enabled from command line.")
                    break
        return new_value

    @classmethod
    def __verify_string_field(
        cls, plugin_instance: RulePlugin, field_name: str, field_value: Any
    ) -> None:
        """
        Verify that the detail field is a valid string.
        """

        if not isinstance(field_value, str):
            raise BadPluginError(
                class_name=type(plugin_instance).__name__, field_name=field_name
            )
        if not field_value:
            raise BadPluginError(
                class_name=type(plugin_instance).__name__,
                field_name=field_name,
                is_empty=True,
            )

    @classmethod
    def __verify_boolean_field(
        cls, plugin_instance: RulePlugin, field_name: str, field_value: Any
    ) -> None:
        """
        Verify that the detail field is a valid boolean.
        """

        if not isinstance(field_value, bool):
            raise BadPluginError(
                class_name=type(plugin_instance).__name__, field_name=field_name
            )

    @classmethod
    def __verify_integer_field(
        cls, plugin_instance: RulePlugin, field_name: str, field_value: Any
    ) -> None:
        """
        Verify that the detail field is a valid integer.
        """

        if not isinstance(field_value, int):
            raise BadPluginError(
                class_name=type(plugin_instance).__name__, field_name=field_name
            )

    def __get_plugin_details(
        self, plugin_instance: RulePlugin, instance_file_name: str
    ) -> FoundPlugin:
        """
        Query the plugin for details and verify that they are reasonable.
        """

        (
            plugin_id,
            plugin_name,
            plugin_description,
            plugin_enabled_by_default,
            plugin_version,
            plugin_interface_version,
            plugin_url,
            plugin_configuration,
            plugin_names,
            plugin_supports_fix,
            plugin_fix_level,
        ) = self.__unpack_plugin_details(plugin_instance)

        self.__verify_string_field(plugin_instance, "plugin_id", plugin_id)
        self.__verify_string_field(plugin_instance, "plugin_name", plugin_name)
        self.__verify_string_field(
            plugin_instance, "plugin_description", plugin_description
        )
        self.__verify_boolean_field(
            plugin_instance, "plugin_enabled_by_default", plugin_enabled_by_default
        )
        self.__verify_string_field(plugin_instance, "plugin_version", plugin_version)
        self.__verify_integer_field(
            plugin_instance, "plugin_interface_version", plugin_interface_version
        )
        if plugin_url:
            self.__verify_string_field(plugin_instance, "plugin_url", plugin_url)
        if plugin_configuration:
            self.__verify_string_field(
                plugin_instance, "plugin_configuration", plugin_configuration
            )
        if plugin_supports_fix:
            self.__verify_integer_field(
                plugin_instance, "plugin_fix_level", plugin_fix_level
            )

        plugin_object = FoundPlugin(
            plugin_id,
            plugin_names,
            plugin_description,
            plugin_instance,
            plugin_enabled_by_default,
            plugin_version,
            plugin_interface_version,
            instance_file_name,
            plugin_url,
            plugin_configuration,
            plugin_supports_fix,
            plugin_fix_level,
            [plugin_id, *plugin_names],
        )

        if plugin_object.plugin_interface_version not in (1, 2, 3):
            raise BadPluginError(
                formatted_message=f"Plugin '{instance_file_name}' with an interface version "
                + f"('{plugin_object.plugin_interface_version}') that is not '1', '2', or '3'."
            )

        return plugin_object

    # pylint: disable=too-many-locals
    def __unpack_plugin_details(self, plugin_instance: RulePlugin) -> Tuple[
        str,
        str,
        str,
        bool,
        str,
        int,
        Optional[str],
        Optional[str],
        List[str],
        bool,
        int,
    ]:
        try:
            instance_details = plugin_instance.get_details()
            plugin_supports_fix = False
            plugin_fix_level = -1
            (
                plugin_id,
                plugin_name,
                plugin_description,
                plugin_enabled_by_default,
                plugin_version,
                plugin_interface_version,
                plugin_url,
                plugin_configuration,
            ) = (
                instance_details.plugin_id,
                instance_details.plugin_name,
                instance_details.plugin_description,
                instance_details.plugin_enabled_by_default,
                instance_details.plugin_version,
                instance_details.plugin_interface_version,
                instance_details.plugin_url,
                instance_details.plugin_configuration,
            )
            if (
                isinstance(instance_details, PluginDetailsV2)
                and plugin_interface_version >= 2
            ):
                plugin_supports_fix = instance_details.plugin_supports_fix
                plugin_fix_level = instance_details.plugin_fix_level
        except Exception as this_exception:
            raise BadPluginError(
                class_name=type(plugin_instance).__name__,
            ) from this_exception

        plugin_id = plugin_id.strip().lower()

        plugin_names = []
        for next_name in plugin_name.lower().split(","):
            if next_name := next_name.strip():
                plugin_names.append(next_name)

        return (
            plugin_id,
            plugin_name,
            plugin_description,
            plugin_enabled_by_default,
            plugin_version,
            plugin_interface_version,
            plugin_url,
            plugin_configuration,
            plugin_names,
            plugin_supports_fix,
            plugin_fix_level,
        )

    # pylint: enable=too-many-locals

    def __register_plugin_id(
        self, plugin_object: FoundPlugin, instance_file_name: str, next_key: str
    ) -> None:
        if not PluginManager.__id_regex.match(next_key):
            raise ValueError(
                f"Unable to register plugin '{instance_file_name}' with id '{next_key}' as "
                + "id is not a valid id in the form 'aannn' or 'aaannn'."
            )

        if next_key in self.__all_ids:
            found_plugin = self.__all_ids[next_key]
            raise ValueError(
                f"Unable to register plugin '{instance_file_name}' with id '{next_key}' as "
                + f"plugin '{found_plugin.plugin_file_name}' is already registered with that id."
            )
        self.__all_ids[next_key] = plugin_object

    def __register_plugin_names(
        self, plugin_object: FoundPlugin, instance_file_name: str
    ) -> None:
        for next_key in plugin_object.plugin_names:
            if not PluginManager.__name_regex.match(next_key):
                raise ValueError(
                    f"Unable to register plugin '{instance_file_name}' with name '{next_key}' as "
                    + "name is not a valid name in the form 'an-an'."
                )
            if next_key in self.__all_ids:
                found_plugin = self.__all_ids[next_key]
                raise ValueError(
                    f"Unable to register plugin '{instance_file_name}' with name '{next_key}' as "
                    + f"plugin '{found_plugin.plugin_file_name}' is already registered with that name."
                )
            self.__all_ids[next_key] = plugin_object

    # pylint: disable=too-many-arguments
    def __register_individual_plugin(
        self,
        plugin_instance: RulePlugin,
        instance_file_name: str,
        command_line_enabled_rules: Set[str],
        command_line_disabled_rules: Set[str],
        properties: ApplicationProperties,
    ) -> None:
        """
        Register an individual plugin for use.
        """

        plugin_object = self.__get_plugin_details(plugin_instance, instance_file_name)

        next_key = plugin_object.plugin_id
        self.__register_plugin_id(plugin_object, instance_file_name, next_key)
        self.__register_plugin_names(plugin_object, instance_file_name)
        if not plugin_object.plugin_description.strip():
            raise ValueError(
                f"Unable to register plugin '{instance_file_name}' with a description string that is blank."
            )
        if not PluginManager.__version_regex.match(plugin_object.plugin_version):
            raise ValueError(
                f"Unable to register plugin '{instance_file_name}' with a version string "
                + "that is not a valid semantic version."
            )

        self.__registered_plugins.append(plugin_object)
        if self.__determine_if_plugin_enabled(
            plugin_object,
            command_line_enabled_rules,
            command_line_disabled_rules,
            properties,
        ):
            self.__enabled_plugins.append(plugin_object)

    # pylint: enable=too-many-arguments

    def __register_plugins(
        self,
        enable_rules_from_command_line: str,
        disable_rules_from_command_line: str,
        properties: ApplicationProperties,
    ) -> None:
        """
        Scan the global namespace for all subclasses of the 'Plugin' class to use as
        plugins.
        """

        command_line_enabled_rules: Set[str] = set()
        command_line_disabled_rules: Set[str] = set()
        self.__registered_plugins = []
        self.__enabled_plugins = []
        self.__all_ids = {}
        if enable_rules_from_command_line:
            for next_rule_identifier in enable_rules_from_command_line.lower().split(
                ","
            ):
                command_line_enabled_rules.add(next_rule_identifier.strip())
        if disable_rules_from_command_line:
            for next_rule_identifier in disable_rules_from_command_line.lower().split(
                ","
            ):
                command_line_disabled_rules.add(next_rule_identifier.strip())

        for plugin_instance, instance_file_name in self.__loaded_classes:
            self.__register_individual_plugin(
                plugin_instance,
                instance_file_name,
                command_line_enabled_rules,
                command_line_disabled_rules,
                properties,
            )

        # Non-windows system may report these in weird orders, so sort them to have
        # a predictable order.
        self.__enabled_plugins = sorted(
            self.__enabled_plugins, reverse=False, key=lambda plugin: plugin.plugin_id
        )

    @property
    def all_plugin_ids(self) -> List[str]:
        """
        Get a list of all plugins by their id.
        """

        return [next_plugin.plugin_id for next_plugin in self.__registered_plugins]

    def is_pragma_on_line(self, line_number: int) -> bool:
        """
        Determine if there is a pragma on the specified line.
        """
        return line_number in self.__pragma_line_numbers

    @classmethod
    def __find_configuration_for_plugin(
        cls,
        next_plugin: FoundPlugin,
        properties: ApplicationProperties,
        always_return_facade: bool = False,
    ) -> Optional[ApplicationPropertiesFacade]:
        plugin_specific_facade, first_facade = None, None
        for next_key_name in next_plugin.plugin_identifiers:
            plugin_section_title = (
                f"{PluginManager.__plugin_prefix}{properties.separator}"
                + f"{next_key_name}{properties.separator}"
            )
            section_facade_candidate = ApplicationPropertiesFacade(
                properties, plugin_section_title
            )
            if not first_facade:
                first_facade = section_facade_candidate
            if section_facade_candidate.property_names:
                plugin_specific_facade = section_facade_candidate
                break

        if always_return_facade and not plugin_specific_facade:
            plugin_specific_facade = first_facade
        return plugin_specific_facade

    def __apply_configuration(
        self, next_plugin: FoundPlugin, properties: ApplicationProperties
    ) -> None:
        try:
            plugin_specific_facade = self.__find_configuration_for_plugin(
                next_plugin, properties, always_return_facade=True
            )

            assert plugin_specific_facade
            next_plugin.plugin_instance.set_configuration_map(plugin_specific_facade)
            next_plugin.plugin_instance.initialize_from_config()
        except Exception as this_exception:
            raise BadPluginError(
                next_plugin.plugin_id,
                inspect.stack()[0].function,
                cause=this_exception,
            ) from this_exception

        if next_plugin.plugin_instance.is_next_token_implemented_in_plugin:
            self.__enabled_plugins_for_next_token.append(next_plugin)
        if next_plugin.plugin_instance.is_next_line_implemented_in_plugin:
            self.__enabled_plugins_for_next_line.append(next_plugin)
        if next_plugin.plugin_instance.is_completed_file_implemented_in_plugin:
            self.__enabled_plugins_for_completed_file.append(next_plugin)
        if next_plugin.plugin_instance.is_starting_new_file_implemented_in_plugin:
            self.__enabled_plugins_for_starting_new_file.append(next_plugin)

    def apply_configuration(
        self, properties: ApplicationProperties, use_full_list: bool = False
    ) -> None:
        """
        Apply any supplied configuration to each of the enabled plugins.
        """

        (
            self.__enabled_plugins_for_starting_new_file,
            self.__enabled_plugins_for_next_token,
            self.__enabled_plugins_for_next_line,
            self.__enabled_plugins_for_completed_file,
        ) = ([], [], [], [])

        proper_list = (
            self.__registered_plugins if use_full_list else self.__enabled_plugins
        )
        for next_plugin in proper_list:
            self.__apply_configuration(next_plugin, properties)

    # pylint: disable=too-many-arguments
    def starting_new_file(
        self,
        file_being_started: str,
        actual_tokens: List[MarkdownToken],
        fix_mode: bool = False,
        temp_output: Optional[TextIOWrapper] = None,
        fix_token_map: Optional[Dict[MarkdownToken, List[FixTokenRecord]]] = None,
        constraint_id_list: Optional[List[str]] = None,
        replace_tokens_list: Optional[List[ReplaceTokensRecord]] = None,
    ) -> PluginScanContext:
        """
        Inform any listeners that a new current file has been started.
        """
        self.__document_pragmas = {}
        self.__document_pragma_ranges = []
        self.__general_pragma_ranges = []
        self.__pragma_line_numbers = []

        for next_plugin in self.__enabled_plugins_for_starting_new_file:
            if constraint_id_list and next_plugin.plugin_id not in constraint_id_list:
                continue
            try:
                next_plugin.plugin_instance.starting_new_file()
            except Exception as this_exception:
                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    cause=this_exception,
                ) from this_exception

        context = PluginScanContext(
            self,
            file_being_started,
            actual_tokens,
            fix_mode,
            temp_output,
            fix_token_map,
            replace_tokens_list,
        )
        context.set_last_line_fixed(None)
        return context

    # pylint: enable=too-many-arguments

    def __completed_file_fix_mode_middle(
        self,
        context: PluginScanContext,
        current_fix_line: Optional[str],
        next_plugin: FoundPlugin,
        line_number: int,
    ) -> Tuple[Optional[str], FixLineRecord]:
        if current_fix_line is not None:
            plugin_id_list = ",".join(next_plugin.plugin_names)
            formatted_message = f"Plugin {next_plugin.plugin_id}({plugin_id_list}) attempted to rewrite a completion line."
            raise BadPluginError(  # pragma: no cover
                formatted_message=formatted_message
            )
        current_fix_line = context.current_fix_line
        line_append_record = FixLineRecord(
            "completed_file", line_number, next_plugin.plugin_id
        )
        return current_fix_line, line_append_record

    def __completed_file_fix_mode_end(
        self,
        context: PluginScanContext,
        current_fix_line: Optional[str],
        line_append_record: Optional[FixLineRecord],
    ) -> None:
        if current_fix_line is not None:
            if self.__show_fix_debug:
                replaced_line = current_fix_line.replace("\n", "\\n").replace(
                    "\t", "\\t"
                )
                print(f"cf-ltw:{replaced_line}:")
            context.file_output.write(current_fix_line)

            assert line_append_record is not None
            context.add_fix_line_record(line_append_record)

    def completed_file(
        self,
        context: PluginScanContext,
        line_number: int,
        context_map: Optional[Dict[str, PluginScanContext]] = None,
    ) -> None:
        """
        Inform any listeners that the current file has been completed.
        """
        context.line_number = line_number
        current_fix_line: Optional[str] = None
        line_append_record: Optional[FixLineRecord] = None

        # This skip added for the assert True.  Without the assert True, code coverage
        # believes that only one of the paths were covered.
        # sourcery skip: remove-assert-true
        for next_plugin in self.__enabled_plugins_for_completed_file:
            if context_map:
                if next_plugin.plugin_id not in context_map:
                    continue
                context = context_map[next_plugin.plugin_id]
            # if context.in_fix_mode and not next_plugin.plugin_supports_fix:
            #     continue
            try:
                if context.in_fix_mode:
                    context.set_current_fix_line(None)
                next_plugin.plugin_instance.completed_file(context)
                if context.in_fix_mode and context.current_fix_line is not None:
                    (
                        current_fix_line,
                        line_append_record,
                    ) = self.__completed_file_fix_mode_middle(
                        context, current_fix_line, next_plugin, line_number
                    )
                assert True
            except Exception as this_exception:
                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    cause=this_exception,
                ) from this_exception
        if context.in_fix_mode:
            self.__completed_file_fix_mode_end(
                context, current_fix_line, line_append_record
            )

    def __next_line_fix_mode_end(
        self,
        context: PluginScanContext,
        line: str,
        is_last_line_in_file: bool,
        was_newline_added_at_end_of_file: bool,
    ) -> None:
        if is_last_line_in_file:
            if self.__show_fix_debug:
                print(
                    f"was_newline_added_at_end_of_file={was_newline_added_at_end_of_file}"
                )
                assert context.last_line_fixed is not None
                replaced_line = context.last_line_fixed.replace("\n", "\\n").replace(
                    "\t", "\\t"
                )
                print(f"fixed:{replaced_line}:")
            is_line_empty = not line
            was_modified = (
                context.last_line_fixed is not None
                and context.last_line_fixed.endswith("\n")
            )
            if self.__show_fix_debug:
                print(f"is_line_empty={is_line_empty}")
                print(f"was_modified={was_modified}")
            line_to_write = line
            was_line_fixed = not (is_line_empty and was_modified)
            # if was_newline_added_at_end_of_file and was_line_fixed:
            #     line_to_write += "\n"
        else:
            line_to_write = line + "\n"
            was_line_fixed = True

        if self.__show_fix_debug:
            replaced_line = line_to_write.replace("\n", "\\n").replace("\t", "\\t")
            print(f"nl-ltw:{replaced_line}:")
        context.file_output.write(line_to_write)
        if was_line_fixed:
            context.set_last_line_fixed(line_to_write)

    def __next_line_fix_mode_before(
        self, context: PluginScanContext, line: str, next_plugin: FoundPlugin
    ) -> None:
        context.set_current_fix_line(None)
        if self.__show_fix_debug:
            replaced_line = line.replace("\n", "\\n").replace("\t", "\\t")
            print(f"{next_plugin.plugin_id}-before:{replaced_line}:")

    def __next_line_fix_mode_after(
        self,
        context: PluginScanContext,
        line: str,
        line_number: int,
        next_plugin: FoundPlugin,
    ) -> str:
        assert context.current_fix_line is not None
        line = context.current_fix_line
        if self.__show_fix_debug:
            replaced_line = line.replace("\n", "\\n").replace("\t", "\\t")
            print(f"{next_plugin.plugin_id}-after :{replaced_line}:")
        line_append_record = FixLineRecord(
            "next_line", line_number, next_plugin.plugin_id
        )
        context.add_fix_line_record(line_append_record)
        return line

    # pylint: disable=too-many-arguments
    def next_line(
        self,
        context: PluginScanContext,
        line_number: int,
        line: str,
        is_last_line_in_file: bool,
        was_newline_added_at_end_of_file: bool,
        context_map: Optional[Dict[str, PluginScanContext]] = None,
    ) -> None:
        """
        Inform any listeners that a new line has been loaded.
        """
        context.line_number = line_number
        for next_plugin in self.__enabled_plugins_for_next_line:
            if context_map:
                if next_plugin.plugin_id not in context_map:
                    continue
                context = context_map[next_plugin.plugin_id]
            # if context.in_fix_mode and not next_plugin.plugin_supports_fix:
            #     continue
            try:
                if context.in_fix_mode:
                    self.__next_line_fix_mode_before(context, line, next_plugin)
                next_plugin.plugin_instance.next_line(context, line)
                if context.current_fix_line is not None:
                    line = self.__next_line_fix_mode_after(
                        context, line, line_number, next_plugin
                    )
            except Exception as this_exception:
                actual_line = line if self.__show_stack_trace else None

                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    line_number=line_number,
                    actual_line=actual_line,
                    cause=this_exception,
                ) from this_exception

        if context.in_fix_mode:
            self.__next_line_fix_mode_end(
                context, line, is_last_line_in_file, was_newline_added_at_end_of_file
            )

    # pylint: enable=too-many-arguments

    def next_token(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        context_map: Optional[Dict[str, PluginScanContext]] = None,
    ) -> None:
        """
        Inform any listeners of a new token that has been processed.
        """
        for next_plugin in self.__enabled_plugins_for_next_token:
            if context_map:
                if next_plugin.plugin_id not in context_map:
                    continue
                context = context_map[next_plugin.plugin_id]

            # if context.in_fix_mode and not next_plugin.plugin_supports_fix:
            #     continue
            try:
                next_plugin.plugin_instance.next_token(context, token)
            except Exception as this_exception:
                actual_token = token if self.__show_stack_trace else None

                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    line_number=token.line_number,
                    column_number=token.column_number,
                    actual_token=actual_token,
                    cause=this_exception,
                ) from this_exception


# pylint: enable=too-many-instance-attributes



"""
Module to provide for the initial process once we have decided to parse a list block.
"""

import logging
from typing import List, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.leaf_blocks.leaf_block_processor_paragraph import (
    LeafBlockProcessorParagraph,
)
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods


class ListBlockPreListHelper:
    """
    Class to provide for the initial process once we have decided to parse a list block.
    """

    # pylint: disable=too-many-arguments
    @staticmethod
    def pre_list(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        marker_width_minus_one: int,
        block_quote_data: BlockQuoteData,
        adj_ws: str,
        position_marker: PositionMarker,
        container_depth: int,
    ) -> Tuple[int, int, int, int, int, List[MarkdownToken], BlockQuoteData]:
        """
        Handle the processing of the first part of the list.
        """
        (
            after_marker_ws_index,
            ws_after_marker,
            ws_before_marker,
            line_to_parse_size,
        ) = ListBlockPreListHelper.__calculate_whitespace_values(
            line_to_parse, start_index, extracted_whitespace
        )

        POGGER.debug("--$--$", start_index, start_index + 1)
        (
            indent_level,
            remaining_whitespace,
            ws_after_marker,
        ) = ListBlockPreListHelper.__calculate_indents(
            after_marker_ws_index,
            line_to_parse_size,
            marker_width_minus_one,
            ws_after_marker,
            ws_before_marker,
            adj_ws,
            container_depth,
        )

        return ListBlockPreListHelper.__check_for_list_nesting(
            parser_state,
            position_marker,
            indent_level,
            after_marker_ws_index,
            block_quote_data,
            remaining_whitespace,
            ws_after_marker,
            ws_before_marker,
        )
        # pylint: enable=too-many-arguments

    @staticmethod
    def __calculate_whitespace_values(
        line_to_parse: str, start_index: int, extracted_whitespace: str
    ) -> Tuple[int, int, int, int]:
        (
            after_marker_ws_index,
            after_marker_whitespace,
        ) = ParserHelper.extract_spaces_verified(line_to_parse, start_index + 1)
        ws_after_marker, ws_before_marker, line_to_parse_size = (
            TabHelper.calculate_length(
                after_marker_whitespace, start_index=start_index + 1
            ),
            TabHelper.calculate_length(extracted_whitespace),
            len(line_to_parse),
        )
        POGGER.debug(
            "after-marker>>$>>total=$", after_marker_whitespace, ws_after_marker
        )
        return (
            after_marker_ws_index,
            ws_after_marker,
            ws_before_marker,
            line_to_parse_size,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __calculate_indents(
        after_marker_ws_index: int,
        line_to_parse_size: int,
        marker_width_minus_one: int,
        ws_after_marker: int,
        ws_before_marker: int,
        adj_ws: str,
        container_depth: int,
    ) -> Tuple[int, int, int]:
        POGGER.debug(
            "--ws_before_marker>>$>>marker_width_minus_one>>$",
            ws_before_marker,
            marker_width_minus_one,
        )
        POGGER.debug("container_depth($)", container_depth)
        POGGER.debug(
            "after_marker_ws_index($) == line_to_parse_size($) and ws_after_marker($)",
            after_marker_ws_index,
            line_to_parse_size,
            ws_after_marker,
        )
        if (
            after_marker_ws_index == line_to_parse_size
            and ws_after_marker
            and not container_depth
        ):
            POGGER.debug("indent1")
            indent_level, remaining_whitespace, ws_after_marker = (
                2 + marker_width_minus_one + len(adj_ws),
                ws_after_marker,
                0,
            )
        else:
            POGGER.debug("indent2")
            if after_marker_ws_index == line_to_parse_size and ws_after_marker == 0:
                ws_after_marker += 1

            indent_level = (
                ws_before_marker + 1 + ws_after_marker + marker_width_minus_one
            )
            if ws_after_marker > 4:
                indent_level, remaining_whitespace, ws_after_marker = (
                    indent_level - ws_after_marker + 1,
                    ws_after_marker - 1,
                    1,
                )
            else:
                remaining_whitespace = 0

        POGGER.debug(
            "ws_after_marker>>$<<indent_level<<$<<rem<<$<<",
            ws_after_marker,
            indent_level,
            remaining_whitespace,
        )
        return indent_level, remaining_whitespace, ws_after_marker

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __check_for_list_nesting(
        parser_state: ParserState,
        position_marker: PositionMarker,
        indent_level: int,
        after_marker_ws_index: int,
        block_quote_data: BlockQuoteData,
        remaining_whitespace: int,
        ws_after_marker: int,
        ws_before_marker: int,
    ) -> Tuple[int, int, int, int, int, List[MarkdownToken], BlockQuoteData]:
        check_list_nesting = True
        if (
            parser_state.token_stack[-1].is_html_block
            or parser_state.token_stack[-1].is_fenced_code_block
        ):
            did_find, _ = LeafBlockProcessorParagraph.check_for_list_in_process(
                parser_state
            )
            if not did_find:
                indent_level = -1
                after_marker_ws_index = -1
                POGGER.debug("BAIL!")
        else:
            POGGER.debug("stack:$:", parser_state.token_stack)
            POGGER.debug("document:$:", parser_state.token_document)
            (
                did_find,
                last_list_index,
            ) = LeafBlockProcessorParagraph.check_for_list_in_process(parser_state)
            if did_find:
                POGGER.debug(
                    "stack[last_list_index]:$:",
                    parser_state.token_stack[last_list_index],
                )
                POGGER.debug(
                    "stack[last_list_index].mmt:$:",
                    parser_state.token_stack[last_list_index].matching_markdown_token,
                )
                check_list_nesting = False

        if check_list_nesting:
            (
                container_level_tokens,
                block_quote_data,
            ) = ListBlockPreListHelper.__handle_list_nesting(
                parser_state, block_quote_data, position_marker
            )
        else:
            container_level_tokens = []
        return (
            indent_level,
            remaining_whitespace,
            ws_after_marker,
            after_marker_ws_index,
            ws_before_marker,
            container_level_tokens,
            block_quote_data,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_list_nesting(
        parser_state: ParserState,
        block_quote_data: BlockQuoteData,
        position_marker: PositionMarker,
    ) -> Tuple[List[MarkdownToken], BlockQuoteData]:
        """
        Resolve any nesting issues with block quotes.
        """
        POGGER.debug(
            ">>block_quote_data.stack_count>>$>>block_quote_data.current_count>>$",
            block_quote_data.stack_count,
            block_quote_data.current_count,
        )
        container_level_tokens: List[MarkdownToken] = []
        adjusted_stack_count = block_quote_data.stack_count
        while block_quote_data.current_count < adjusted_stack_count:
            assert (
                not container_level_tokens
            ), "Container tokens cannot have been filled."
            last_block_index = parser_state.find_last_block_quote_on_stack()
            previous_last_block_token = cast(
                BlockQuoteMarkdownToken,
                parser_state.token_stack[last_block_index].matching_markdown_token,
            )
            POGGER.debug(
                "last_block_index>>$-->$", last_block_index, previous_last_block_token
            )
            container_level_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                until_this_index=last_block_index,
                include_block_quotes=True,
                include_lists=True,
            )
            POGGER.debug("container_level_tokens>>$", container_level_tokens)
            POGGER.debug("stack>>$", parser_state.token_stack)
            last_markdown_token = parser_state.token_stack[-1].matching_markdown_token
            POGGER.debug("last token>>$", last_markdown_token)
            POGGER.debug("position_marker.line_number>>$", position_marker.line_number)
            last_block_index = parser_state.find_last_block_quote_on_stack()
            POGGER.debug("last_block_index>>$", last_block_index)
            POGGER.debug(
                "parser_state.token_stack[-1]>>$", parser_state.token_stack[-1]
            )
            POGGER.debug(
                "last_markdown_token>>$",
                last_markdown_token,
            )

            first_conditional = not last_markdown_token
            second_conditional = (
                (not first_conditional)
                and last_markdown_token
                and (position_marker.line_number == last_markdown_token.line_number)
            )
            third_conditional = parser_state.token_stack[-1].is_block_quote

            secondary_conditionals = (
                first_conditional or second_conditional or third_conditional
            )
            POGGER.debug(
                "secondary_conditionals>>$ = first_conditional:$ or second_conditional:$ or "
                + "third_conditional:$",
                secondary_conditionals,
                first_conditional,
                second_conditional,
                third_conditional,
            )
            all_conditionals = last_block_index and secondary_conditionals
            POGGER.debug(
                "all_conditionals>>$ = last_block_index:$ or a2:$",
                all_conditionals,
                last_block_index,
                secondary_conditionals,
            )
            if all_conditionals:
                ListBlockPreListHelper.__handle_list_nesting_all_conditionals(
                    parser_state, last_block_index, previous_last_block_token
                )

            adjusted_stack_count -= 1

        if adjusted_stack_count != block_quote_data.stack_count:
            block_quote_data = BlockQuoteData(
                block_quote_data.current_count, adjusted_stack_count
            )
        return container_level_tokens, block_quote_data

    @staticmethod
    def __handle_list_nesting_all_conditionals(
        parser_state: ParserState,
        last_block_index: int,
        previous_last_block_token: BlockQuoteMarkdownToken,
    ) -> None:
        current_last_block_token = cast(
            BlockQuoteMarkdownToken,
            parser_state.token_stack[last_block_index].matching_markdown_token,
        )
        POGGER.debug(
            "last_block_index>>$-->$",
            last_block_index,
            current_last_block_token,
        )

        POGGER.debug(
            "prev>>$<<, current>>$<<",
            previous_last_block_token,
            current_last_block_token,
        )
        removed_leading_spaces = previous_last_block_token.remove_last_bleading_space()
        POGGER.debug("removed_leading_spaces>>$<<", removed_leading_spaces)
        assert removed_leading_spaces is not None, "Removed spaces cannot be None."
        POGGER.debug(
            "prev>>$<<, current>>$<<",
            previous_last_block_token,
            current_last_block_token,
        )
        POGGER.debug(
            "__handle_list_nesting_all_conditionals>>block_token>>$",
            current_last_block_token,
        )
        current_last_block_token.add_bleading_spaces(
            removed_leading_spaces, skip_adding_newline=False
        )
        POGGER.debug(
            "__handle_list_nesting_all_conditionals>>block_token>>$",
            current_last_block_token,
        )
        POGGER.debug(
            "prev>>$<<, current>>$<<",
            previous_last_block_token,
            current_last_block_token,
        )


# pylint: enable=too-few-public-methods



"""
Module to provide a tokenization of a markdown-encoded string.
"""

import logging
import os
from typing import List, Optional, Tuple, cast

from application_properties import ApplicationProperties

from pymarkdown.coalesce.coalesce_processor import CoalesceProcessor
from pymarkdown.container_blocks.container_block_processor import (
    ContainerBlockProcessor,
)
from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.extension_manager.extension_manager import ExtensionManager
from pymarkdown.extensions.front_matter_extension import FrontMatterExtension
from pymarkdown.extensions.pragma_token import PragmaToken
from pymarkdown.general.bad_tokenization_error import BadTokenizationError
from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.requeue_line_info import RequeueLineInfo
from pymarkdown.general.source_providers import InMemorySourceProvider, SourceProvider
from pymarkdown.html.html_helper import HtmlHelper
from pymarkdown.inline.inline_character_reference_helper import (
    InlineCharacterReferenceHelper,
)
from pymarkdown.inline.inline_processor import InlineProcessor
from pymarkdown.leaf_blocks.leaf_block_helper import LeafBlockHelper
from pymarkdown.leaf_blocks.leaf_block_processor_paragraph import (
    LeafBlockProcessorParagraph,
)
from pymarkdown.leaf_blocks.table_block_processor import TableBlockHelper
from pymarkdown.links.link_parse_helper import LinkParseHelper
from pymarkdown.links.link_reference_definition_helper import (
    LinkReferenceDefinitionHelper,
)
from pymarkdown.tokens.blank_line_markdown_token import BlankLineMarkdownToken
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.end_of_stream_token import EndOfStreamToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import (
    BlockQuoteStackToken,
    DocumentStackToken,
    ParagraphStackToken,
    StackToken,
)

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-many-lines


class TokenizedMarkdown:
    """
    Class to provide a tokenization of a markdown-encoded string.
    """

    def __init__(self, resource_path: Optional[str] = None) -> None:
        """
        Initializes a new instance of the TokenizedMarkdown class.
        """

        self.__tokenized_document: List[MarkdownToken] = []
        self.__token_stack: List[StackToken] = []
        self.__parse_properties: Optional[ParseBlockPassProperties] = None
        self.__source_provider: Optional[SourceProvider] = None
        self.__extension_manager: Optional[ExtensionManager] = None

        if not resource_path:
            resource_path = os.path.join(os.path.split(__file__)[0], "..", "resources")
            resource_path = os.path.abspath(resource_path)
        InlineCharacterReferenceHelper.initialize(resource_path)

    def apply_configuration(
        self,
        application_properties: ApplicationProperties,
        extension_manager: ExtensionManager,
    ) -> None:
        """
        Apply any configuration map.
        """
        _ = application_properties
        self.__extension_manager = extension_manager
        self.__parse_properties = ParseBlockPassProperties(extension_manager)

    def transform_from_provider(
        self,
        source_provider: Optional[SourceProvider],
        do_add_end_of_stream_token: bool = False,
    ) -> List[MarkdownToken]:
        """
        Transform the data from the source provider into a Markdown token stream.
        """
        self.__source_provider = source_provider
        return self.__transform(do_add_end_of_stream_token)

    def transform(
        self,
        your_text_string: str,
        show_debug: bool = False,
        do_add_end_of_stream_token: bool = False,
    ) -> List[MarkdownToken]:
        """
        Transform a text string in a Markdown format into a Markdown token stream.
        This function should only be used as a simplified manner of accessing the
        functionality for the purposes of testing.
        """

        logging.getLogger().setLevel(logging.DEBUG if show_debug else logging.WARNING)
        ParserLogger.sync_on_next_call()
        self.__source_provider = InMemorySourceProvider(your_text_string)
        return self.__transform(do_add_end_of_stream_token)

    def __transform(self, do_add_end_of_stream_token: bool) -> List[MarkdownToken]:
        """
        Transform a markdown-encoded string into an array of tokens.
        """
        try:
            self.__tokenized_document = []
            self.__token_stack = []

            assert (
                self.__extension_manager is not None
            ), "Extension manager must be initialized by this point."
            InlineProcessor.initialize(self.__extension_manager)
            LinkParseHelper.initialize()

            POGGER.debug("\n\n>>>>>>>parse_blocks_pass>>>>>>")
            first_pass_results = self.__parse_blocks_pass(do_add_end_of_stream_token)

            POGGER.debug("\n\n>>>>>>>coalesce_text_blocks>>>>>>")
            coalesced_results = CoalesceProcessor.coalesce_text_blocks(
                first_pass_results
            )

            POGGER.debug("\n\n>>>>>>>parse_inline>>>>>>")
            assert (
                self.__parse_properties is not None
            ), "Parse properties must be defined by this point."
            final_pass_results = InlineProcessor.parse_inline(
                coalesced_results, self.__parse_properties
            )

            final_coalesced_results = CoalesceProcessor.coalesce_text_blocks(
                final_pass_results, only_change_text_blocks=True
            )
            POGGER.debug("\n\n>>>>>>>final_pass_results>>>>>>")
            return final_coalesced_results
        except Exception as this_exception:
            raise BadTokenizationError(
                "An unhandled error occurred processing the document."
            ) from this_exception

    def __parse_blocks_pass(
        self, do_add_end_of_stream_token: bool
    ) -> List[MarkdownToken]:
        """
        The first pass at the tokens is to deal with blocks.
        """

        assert (
            self.__parse_properties is not None
        ), "Parse properties must be defined by this point."
        assert (
            self.__source_provider is not None
        ), "Source provider must be defined by this point."
        self.__token_stack = [DocumentStackToken()]
        self.__tokenized_document = []
        self.__parse_properties.pragma_lines = {}

        POGGER.debug("---")
        try:
            first_line_in_document, line_number = (
                self.__source_provider.get_next_line(),
                1,
            )
            POGGER.debug("---$---", first_line_in_document)
            (
                first_line_in_document,
                line_number,
                requeue,
            ) = self.__process_front_matter_header_if_present(
                first_line_in_document, line_number, []
            )
            (
                did_start_close,
                did_started_close,
                keep_on_going,
                ignore_link_definition_start,
                ignore_table_start,
            ) = (first_line_in_document is None, False, True, False, False)
            next_line_in_document = first_line_in_document
            while keep_on_going:
                (
                    keep_on_going,
                    did_start_close,
                    did_started_close,
                    ignore_link_definition_start,
                    ignore_table_start,
                    requeue,
                    line_number,
                    next_line_in_document,
                ) = self.__parse_blocks_pass_next_line(
                    next_line_in_document,
                    line_number,
                    requeue,
                    did_start_close,
                    did_started_close,
                    ignore_link_definition_start,
                    ignore_table_start,
                )
        except AssertionError as this_exception:
            error_message = f"A project assertion failed on line {line_number} of the current document."
            raise BadTokenizationError(error_message) from this_exception

        if do_add_end_of_stream_token:
            self.__tokenized_document.append(EndOfStreamToken(line_number))

        if self.__parse_properties.pragma_lines:
            self.__tokenized_document.append(
                PragmaToken(self.__parse_properties.pragma_lines)
            )
        return self.__tokenized_document

    # pylint: disable=too-many-arguments
    def __parse_blocks_pass_next_line(
        self,
        next_line_in_document: Optional[str],
        line_number: int,
        requeue: List[str],
        did_start_close: bool,
        did_started_close: bool,
        ignore_link_definition_start: bool,
        ignore_table_start: bool,
    ) -> Tuple[bool, bool, bool, bool, bool, List[str], int, Optional[str]]:
        POGGER.debug("next-line>>$", next_line_in_document)
        POGGER.debug("stack>>$", self.__token_stack)
        POGGER.debug("current_block>>$", self.__token_stack[-1])
        POGGER.debug("line_number>>$", line_number)
        POGGER.debug("---")

        assert (
            self.__parse_properties is not None
        ), "Parse properties must be defined by this point."
        parser_state = ParserState(
            self.__token_stack,
            self.__tokenized_document,
            TokenizedMarkdown.__close_open_blocks,
            self.__handle_blank_line,
            self.__parse_properties,
        )
        keep_on_going = True
        if did_start_close:
            (
                did_started_close,
                did_start_close,
                tokens_from_line,
                line_number,
                keep_on_going,
                requeue_line_info,
            ) = self.__main_pass_did_start_close(parser_state, line_number)
        else:
            assert (
                next_line_in_document is not None
            ), "If not closing, must have a next line to process."
            position_marker = PositionMarker(line_number, 0, next_line_in_document)
            (
                tfl,
                requeue_line_info,
            ) = self.__main_pass_did_not_start_close(
                parser_state,
                position_marker,
                next_line_in_document,
                ignore_link_definition_start,
                ignore_table_start,
            )
            assert tfl is not None, "Markdown token list cannot be empty."
            tokens_from_line = tfl

        if keep_on_going:
            (
                line_number,
                ignore_link_definition_start,
                ignore_table_start,
                next_line_in_document,
                did_start_close,
                did_started_close,
            ) = self.__main_pass_keep_on_going(
                line_number,
                requeue_line_info,
                requeue,
                tokens_from_line,
                did_start_close,
                did_started_close,
            )
        return (
            keep_on_going,
            did_start_close,
            did_started_close,
            ignore_link_definition_start,
            ignore_table_start,
            requeue,
            line_number,
            next_line_in_document,
        )

    # pylint: enable=too-many-arguments

    def __main_pass_did_start_close(
        self, parser_state: ParserState, line_number: int
    ) -> Tuple[bool, bool, List[MarkdownToken], int, bool, Optional[RequeueLineInfo]]:
        POGGER.debug("\n\ncleanup")

        was_link_definition_started_before_close = self.__token_stack[
            -1
        ].was_link_definition_started
        was_table_started_before_close = self.__token_stack[-1].was_table_block_started

        (
            tokens_from_line,
            requeue_line_info,
        ) = TokenizedMarkdown.__close_open_blocks(
            parser_state,
            self.__tokenized_document,
            include_block_quotes=True,
            include_lists=True,
            caller_can_handle_requeue=True,
            was_forced=True,
        )
        if not self.__tokenized_document:
            self.__tokenized_document.extend(tokens_from_line)

        keep_on_going = bool(requeue_line_info and requeue_line_info.lines_to_requeue)
        did_start_close = not keep_on_going
        if keep_on_going:
            assert (
                was_link_definition_started_before_close
                or was_table_started_before_close
            ), "LRD parsing or Table parsing must have been started."
            assert (
                requeue_line_info is not None
                and not requeue_line_info.lines_to_requeue[0]
            ), "Cannot handling requeing"

            del requeue_line_info.lines_to_requeue[0]
            line_number -= 1

            tokens_from_line = []
        return (
            True,
            did_start_close,
            tokens_from_line,
            line_number,
            keep_on_going,
            requeue_line_info,
        )

    # pylint: disable=too-many-arguments
    def __main_pass_did_not_start_close(
        self,
        parser_state: ParserState,
        position_marker: PositionMarker,
        next_line_in_document: str,
        ignore_link_definition_start: bool,
        ignore_table_start: bool,
    ) -> Tuple[Optional[List[MarkdownToken]], Optional[RequeueLineInfo]]:
        POGGER.debug(">>>>$", self.__tokenized_document)

        if not next_line_in_document or not next_line_in_document.strip(
            Constants.ascii_whitespace
        ):
            POGGER.debug("call __parse_blocks_pass>>handle_blank_line")
            (
                tokens_from_line,
                requeue_line_info,
            ) = self.__handle_blank_line(
                parser_state,
                next_line_in_document,
                from_main_transform=True,
                position_marker=position_marker,
            )
        else:
            POGGER.debug("\n\nnormal lines")
            assert (
                self.__parse_properties is not None
            ), "Parse properties must be defined by this point."
            (
                tokens_from_line,
                _,
                _,
                requeue_line_info,
                _,
                _,
            ) = ContainerBlockProcessor.parse_line_for_container_blocks(
                parser_state,
                position_marker,
                ignore_link_definition_start,
                ignore_table_start,
                self.__parse_properties,
                0,
            )

        POGGER.debug("<<<<$", self.__tokenized_document)

        return tokens_from_line, requeue_line_info

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def __main_pass_keep_on_going(
        self,
        line_number: int,
        requeue_line_info: Optional[RequeueLineInfo],
        requeue: List[str],
        tokens_from_line: List[MarkdownToken],
        did_start_close: bool,
        did_started_close: bool,
    ) -> Tuple[int, bool, bool, Optional[str], bool, bool]:
        (line_number, ignore_link_definition_start, ignore_table_start) = (
            self.__handle_parse_increment_line(line_number, requeue_line_info, requeue)
        )

        POGGER.debug(
            "---\nbefore>>$",
            self.__tokenized_document,
        )
        POGGER.debug("before>>$", tokens_from_line)
        self.__tokenized_document.extend(tokens_from_line)
        POGGER.debug(
            "after>>$",
            self.__tokenized_document,
        )
        if requeue:
            POGGER.debug("requeue>>$", requeue)
        POGGER.debug("---")

        (
            next_line_in_document,
            did_start_close,
            did_started_close,
        ) = self.__determine_next_line_to_process(
            requeue, did_start_close, did_started_close
        )
        return (
            line_number,
            ignore_link_definition_start,
            ignore_table_start,
            next_line_in_document,
            did_start_close,
            did_started_close,
        )

    # pylint: enable=too-many-arguments

    def __handle_parse_increment_line(
        self,
        line_number: int,
        requeue_line_info: Optional[RequeueLineInfo],
        requeue: List[str],
    ) -> Tuple[int, bool, bool]:
        if requeue_line_info:
            number_of_lines_to_requeue = len(requeue_line_info.lines_to_requeue)
            POGGER.debug("\n\n---lines_to_requeue>>$", number_of_lines_to_requeue)
            assert len(requeue_line_info.lines_to_requeue) > 0
            assert self.__parse_properties is not None

            requeue.insert(0, requeue_line_info.lines_to_requeue[0])
            del requeue_line_info.lines_to_requeue[0]
            c = number_of_lines_to_requeue - 1
            while c > 0:
                line_number -= 1
                while line_number in self.__parse_properties.pragma_lines:
                    x = self.__parse_properties.pragma_lines[line_number]
                    del self.__parse_properties.pragma_lines[line_number]
                    requeue.insert(0, x)
                    line_number -= 1
                x = requeue_line_info.lines_to_requeue[0]
                del requeue_line_info.lines_to_requeue[0]
                requeue.insert(0, x)
                c -= 1
            # for i in requeue_line_info.lines_to_requeue:
            #     requeue.insert(0, i)
            ignore_link_definition_start = requeue_line_info.force_ignore_first_as_lrd
            ignore_table_start = requeue_line_info.force_ignore_first_as_table
        else:
            ignore_link_definition_start = False
            ignore_table_start = False
            line_number += 1
        POGGER.debug("line_number>>$\n---", line_number)

        return line_number, ignore_link_definition_start, ignore_table_start

    def __determine_next_line_to_process(
        self, requeue: List[str], did_start_close: bool, did_started_close: bool
    ) -> Tuple[Optional[str], bool, bool]:
        """
        For the parse_blocks_pass function, determine the next token to parse.
        """

        next_line_in_document: Optional[str] = None
        if requeue:
            POGGER.debug(">>Requeues present")
            next_line_in_document = requeue[0]
            del requeue[0]
            POGGER.debug(">>Requeue>>$", next_line_in_document)
            POGGER.debug(">>Requeues left>>$", requeue)
        elif did_started_close:
            did_start_close = True
        else:
            assert (
                self.__source_provider is not None
            ), "Source provider must be defined by this point."
            next_line_in_document = self.__source_provider.get_next_line()
            if next_line_in_document is None:
                did_start_close = True

        return next_line_in_document, did_start_close, did_started_close

    @staticmethod
    def __close_open_blocks_log_args(
        include_block_quotes: bool,
        include_lists: bool,
        until_this_index: int,
        caller_can_handle_requeue: bool,
        was_forced: bool,
    ) -> None:
        if include_block_quotes:
            POGGER.debug("cob-include_block_quotes>>$", include_block_quotes)
        if include_lists:
            POGGER.debug("cob-include_lists>>$", include_lists)
        if until_this_index != -1:
            POGGER.debug("cob-until_this_index>>$", until_this_index)
        if caller_can_handle_requeue:
            POGGER.debug("cob-caller_can_handle_requeue>>$", caller_can_handle_requeue)
        if was_forced:
            POGGER.debug("cob-was_forced>>$", was_forced)

    # pylint: disable=too-many-arguments
    @staticmethod
    def __close_open_blocks(  # noqa: C901
        parser_state: ParserState,
        destination_array: Optional[List[MarkdownToken]] = None,
        only_these_blocks: Optional[List[type]] = None,
        include_block_quotes: bool = False,
        include_lists: bool = False,
        until_this_index: int = -1,
        caller_can_handle_requeue: bool = False,
        requeue_reset: bool = False,
        was_forced: bool = False,
    ) -> Tuple[List[MarkdownToken], Optional[RequeueLineInfo]]:
        """
        Close any open blocks that are currently on the stack.
        """

        requeue_line_info = None
        new_tokens = destination_array or []
        POGGER.debug("cob-start>>$", parser_state.token_stack)
        POGGER.debug(
            "cob-start>>$",
            parser_state.token_document,
        )
        if only_these_blocks:
            POGGER.debug("cob-only_these_blocks>>$", only_these_blocks)
        TokenizedMarkdown.__close_open_blocks_log_args(
            include_block_quotes,
            include_lists,
            until_this_index,
            caller_can_handle_requeue,
            was_forced,
        )
        while not parser_state.token_stack[-1].is_document:
            was_close_forced = was_forced

            (
                can_continue,
                adjusted_tokens,
                requeue_line_info,
                was_close_forced,
            ) = TokenizedMarkdown.__process_next_close_element(
                parser_state,
                was_close_forced,
                caller_can_handle_requeue,
                until_this_index,
                include_lists,
                include_block_quotes,
                only_these_blocks,
            )
            new_tokens.extend(adjusted_tokens)
            if not can_continue:
                break

        POGGER.debug("cob-end>>$", parser_state.token_stack)
        POGGER.debug("cob-end>>$", parser_state.token_document)
        POGGER.debug("cob-end>>new_tokens>>$", new_tokens)
        if caller_can_handle_requeue:
            TokenizedMarkdown.__log_requeue(
                requeue_line_info, requeue_reset, parser_state
            )
        return new_tokens, requeue_line_info

    # pylint: enable=too-many-arguments
    @staticmethod
    def __log_requeue(
        requeue_line_info: Optional[RequeueLineInfo],
        requeue_reset: bool,
        parser_state: ParserState,
    ) -> None:
        if not requeue_line_info:
            POGGER.debug("requeue_line_info>>no requeue")
            return
        POGGER.debug(
            "cob>>lines_to_requeue>>$",
            requeue_line_info.lines_to_requeue,
        )
        if requeue_reset:
            assert (
                parser_state.original_line_to_parse is not None
            ), "Original line to parse must be defined by this point."
            POGGER.debug(
                "cob>>original_line_to_parse>$>", parser_state.original_line_to_parse
            )
            # assert not requeue_line_info.lines_to_requeue[
            #     0
            # ], "Resetting implies something must have been there to reset."
            requeue_line_info.lines_to_requeue[0] = parser_state.original_line_to_parse
            POGGER.debug(
                "cob>>(adjusted)lines_to_requeue>>$",
                requeue_line_info.lines_to_requeue,
            )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __process_next_close_element(
        parser_state: ParserState,
        was_close_forced: bool,
        caller_can_handle_requeue: bool,
        until_this_index: int,
        include_lists: bool,
        include_block_quotes: bool,
        only_these_blocks: Optional[List[type]],
    ) -> Tuple[bool, List[MarkdownToken], Optional[RequeueLineInfo], bool]:
        requeue_line_info = None

        POGGER.debug("cob>>$", parser_state.token_stack)
        can_continue, was_close_forced = TokenizedMarkdown.__can_close_continue(
            parser_state,
            only_these_blocks,
            include_block_quotes,
            include_lists,
            until_this_index,
            was_close_forced,
        )
        POGGER.debug("can_continue>>$", can_continue)
        POGGER.debug("was_close_forced>>$", was_close_forced)
        if can_continue:
            if parser_state.token_stack[-1].was_link_definition_started:
                (
                    can_continue,
                    adjusted_tokens,
                    requeue_line_info,
                ) = TokenizedMarkdown.__close_open_blocks_lrd(
                    parser_state, caller_can_handle_requeue
                )
            elif parser_state.token_stack[-1].was_table_block_started:
                (
                    can_continue,
                    adjusted_tokens,
                    requeue_line_info,
                ) = TokenizedMarkdown.__close_open_blocks_table(
                    parser_state, caller_can_handle_requeue
                )
            else:
                adjusted_tokens = TokenizedMarkdown.__close_open_blocks_normal(
                    parser_state, was_close_forced
                )
        else:
            adjusted_tokens = []
        return can_continue, adjusted_tokens, requeue_line_info, was_close_forced

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __can_close_continue(
        parser_state: ParserState,
        only_these_blocks: Optional[List[type]],
        include_block_quotes: bool,
        include_lists: bool,
        until_this_index: int,
        was_close_forced: bool,
    ) -> Tuple[bool, bool]:
        can_continue = True
        if only_these_blocks:
            POGGER.debug("cob-only-type>>$", only_these_blocks)
            POGGER.debug("cob-only-type>>$", type(parser_state.token_stack[-1]))
            can_continue = type(parser_state.token_stack[-1]) in only_these_blocks
            if not can_continue:
                POGGER.debug("cob>>not in only")
        elif not include_block_quotes and parser_state.token_stack[-1].is_block_quote:
            POGGER.debug("cob>>not block quotes")
            can_continue = False
        elif not include_lists and parser_state.token_stack[-1].is_list:
            POGGER.debug("cob>>not lists")
            can_continue = False
        elif until_this_index != -1:
            token_stack_size = len(parser_state.token_stack)
            POGGER.debug(
                "NOT ME!!!!$<<$<<",
                until_this_index,
                token_stack_size,
            )
            if until_this_index >= token_stack_size:
                can_continue = False
            else:
                was_close_forced = True

        POGGER.debug("cob>>can_continue=$", can_continue)
        POGGER.debug("cob>>was_close_forced=$", was_close_forced)
        return can_continue, was_close_forced

    # pylint: enable=too-many-arguments

    @staticmethod
    def __close_open_blocks_normal(
        parser_state: ParserState, was_close_forced: bool
    ) -> List[MarkdownToken]:
        POGGER.debug(
            "cob-rem>>$",
            parser_state.token_document,
        )
        adjusted_tokens = TokenizedMarkdown.__remove_top_element_from_stack(
            parser_state, was_close_forced
        )
        POGGER.debug(
            "cob-rem<<$",
            parser_state.token_document,
        )
        POGGER.debug("cob-adj<<$", adjusted_tokens)
        return adjusted_tokens

    @staticmethod
    def __close_open_blocks_lrd(
        parser_state: ParserState, caller_can_handle_requeue: bool
    ) -> Tuple[bool, List[MarkdownToken], Optional[RequeueLineInfo]]:
        POGGER.debug("cob->process_link_reference_definition>>stopping link definition")
        empty_position_marker = PositionMarker(-1, 0, "")
        (
            outer_processed,
            did_complete_lrd,
            did_pause_lrd,
            requeue_line_info,
            adjusted_tokens,
        ) = LinkReferenceDefinitionHelper.process_link_reference_definition(
            parser_state, empty_position_marker, "", "", "", 0, 0, ""
        )

        POGGER.debug("BOOOM: caller_can_handle_requeue=$", caller_can_handle_requeue)
        can_continue = not (
            caller_can_handle_requeue
            and requeue_line_info
            and requeue_line_info.lines_to_requeue
        )
        if not can_continue:
            POGGER.debug("BOOOM-->break")
        else:
            assert not (
                requeue_line_info and requeue_line_info.lines_to_requeue
            ), "Cannot requeue at this point."
            POGGER.debug(
                "cob->process_link_reference_definition>>outer_processed>>$",
                outer_processed,
            )
            POGGER.debug(
                "cob->process_link_reference_definition>>did_complete_lrd>>$",
                did_complete_lrd,
            )
            POGGER.debug(
                "cob->process_link_reference_definition>>adjusted_tokens>>$",
                adjusted_tokens,
            )
            assert not did_pause_lrd, "LRD parsing must not be paused."
        return can_continue, adjusted_tokens, requeue_line_info

    @staticmethod
    def __close_open_blocks_table(
        parser_state: ParserState, caller_can_handle_requeue: bool
    ) -> Tuple[bool, List[MarkdownToken], Optional[RequeueLineInfo]]:
        POGGER.debug("cob->process_table>>stopping table")
        empty_position_marker = PositionMarker(-1, 0, "")
        (
            outer_processed,
            did_complete_table,
            did_pause_table,
            requeue_line_info,
            adjusted_tokens,
        ) = TableBlockHelper.process_table_rows(
            parser_state, empty_position_marker, "", "", "", 0, 0, ""
        )

        POGGER.debug("BOOOM: caller_can_handle_requeue=$", caller_can_handle_requeue)
        can_continue = not (
            caller_can_handle_requeue
            and requeue_line_info
            and requeue_line_info.lines_to_requeue
        )
        if not can_continue:
            POGGER.debug("BOOOM-->break")
        else:
            assert not (
                requeue_line_info and requeue_line_info.lines_to_requeue
            ), "Cannot requeue at this point."
            POGGER.debug(
                "cob->process_table>>outer_processed>>$",
                outer_processed,
            )
            POGGER.debug(
                "cob->process_table>>did_complete_table>>$",
                did_complete_table,
            )
            POGGER.debug(
                "cob->process_table>>adjusted_tokens>>$",
                adjusted_tokens,
            )
            assert not did_pause_table, "Table parsing must not be paused."
        return can_continue, adjusted_tokens, requeue_line_info

    @staticmethod
    def __remove_top_element_from_stack(
        parser_state: ParserState, was_forced: bool
    ) -> List[MarkdownToken]:
        """
        Once it is decided that we need to remove the top element from the stack,
        make sure to do it uniformly.
        """

        POGGER.debug("cob->top_element->$", parser_state.token_stack[-1])
        POGGER.debug("cob->was_forced->$", was_forced)
        extra_elements = []
        extra_end_data = None
        if parser_state.token_stack[-1].is_indented_code_block:
            extra_elements.extend(
                LeafBlockHelper.extract_markdown_tokens_back_to_blank_line(
                    parser_state, was_forced
                )
            )
        elif parser_state.token_stack[-1].is_fenced_code_block:
            extra_end_data = ":"

        new_tokens = [
            parser_state.token_stack[-1].generate_close_markdown_token_from_stack_token(
                was_forced=was_forced, extra_end_data=extra_end_data
            ),
            *extra_elements,
        ]

        del parser_state.token_stack[-1]
        return new_tokens

    @staticmethod
    def __handle_blank_line_init(
        from_main_transform: bool, input_line: str
    ) -> Tuple[Optional[List[type]], bool, int, str]:
        do_include_block_quotes = from_main_transform
        close_only_these_blocks: Optional[List[type]] = (
            None if from_main_transform else [ParagraphStackToken]
        )

        POGGER.debug("hbl>>from_main_transform>>$", from_main_transform)
        POGGER.debug("hbl>>close_only_these_blocks>>$", close_only_these_blocks)
        POGGER.debug("hbl>>do_include_block_quotes>>$", do_include_block_quotes)

        (
            non_whitespace_index,
            extracted_whitespace,
        ) = ParserHelper.extract_ascii_whitespace_verified(input_line, 0)
        return (
            close_only_these_blocks,
            do_include_block_quotes,
            non_whitespace_index,
            extracted_whitespace,
        )

    @staticmethod
    def __handle_blank_line(
        parser_state: ParserState,
        input_line: str,
        from_main_transform: bool,
        position_marker: Optional[PositionMarker] = None,
    ) -> Tuple[Optional[List[MarkdownToken]], Optional[RequeueLineInfo]]:
        """
        Handle the processing of a blank line.
        """
        assert position_marker is not None
        parser_state.mark_start_information(position_marker)

        (
            close_only_these_blocks,
            do_include_block_quotes,
            non_whitespace_index,
            extracted_whitespace,
        ) = TokenizedMarkdown.__handle_blank_line_init(from_main_transform, input_line)

        (
            new_tokens,
            force_default_handling,
            requeue_line_info,
        ) = TokenizedMarkdown.__handle_blank_line_token_stack(parser_state)
        if requeue_line_info:
            return new_tokens, requeue_line_info

        if from_main_transform:
            POGGER.debug("hbl>>__handle_blank_line_in_block_quote")
            TokenizedMarkdown.__handle_blank_line_in_block_quote(
                parser_state, new_tokens
            )
            new_tokens = TokenizedMarkdown.__handle_blank_line_close_first_quote(
                parser_state, new_tokens
            )

        if force_default_handling or new_tokens is None:
            POGGER.debug("hbl>>default blank handling-->cob")
            n_tokens, _ = TokenizedMarkdown.__close_open_blocks(
                parser_state,
                only_these_blocks=close_only_these_blocks,
                include_block_quotes=do_include_block_quotes,
                was_forced=True,
            )
            if new_tokens:
                new_tokens.extend(n_tokens)
            else:
                new_tokens = n_tokens

        list_stack_index = parser_state.find_last_list_block_on_stack()
        block_stack_index = parser_state.find_last_block_quote_on_stack()
        POGGER.debug("list_stack_index>>$", list_stack_index)
        POGGER.debug("block_stack_index>>$", block_stack_index)
        if list_stack_index > 0 and list_stack_index > block_stack_index:

            TokenizedMarkdown.__handle_blank_line_bravo(parser_state, block_stack_index)

            list_token = cast(
                ListStartMarkdownToken,
                parser_state.token_stack[list_stack_index].matching_markdown_token,
            )
            POGGER.debug("hbl>>last_block_token>>$", list_token)
            list_token.add_leading_spaces("")
            POGGER.debug("hbl>>last_block_token>>$", list_token)

        POGGER.debug("hbl>>new_tokens>>$", new_tokens)
        assert non_whitespace_index == len(
            input_line
        ), "Index must be set to the end of the line, since processing has completed."
        new_tokens.append(BlankLineMarkdownToken(extracted_whitespace, position_marker))
        POGGER.debug("hbl>>new_tokens>>$", new_tokens)

        return new_tokens, None

    @staticmethod
    def __handle_blank_line_bravo(
        parser_state: ParserState, block_stack_index: int
    ) -> None:
        search_index = 0
        next_index = block_stack_index - 1
        while next_index > 0:
            if parser_state.token_stack[next_index].is_list:
                search_index = next_index
                break
            next_index -= 1
        if search_index:
            found_markdown_token = parser_state.token_stack[
                next_index
            ].matching_markdown_token
            assert found_markdown_token is not None
            block_copy_token = next(  # pragma: no cover
                (
                    j
                    for j in parser_state.block_copy
                    if (
                        j is not None
                        and j.line_number == found_markdown_token.line_number
                        and j.column_number == found_markdown_token.column_number
                    )
                ),
                None,
            )

            # assert on next line was if statement with rest of function in if block
            assert block_copy_token is not None
            assert found_markdown_token.is_list_start
            found_markdown_list_token = cast(
                ListStartMarkdownToken, found_markdown_token
            )
            found_markdown_token_leading_spaces = (
                found_markdown_list_token.leading_spaces
            )
            assert block_copy_token.is_list_start
            block_copy_list_token = cast(ListStartMarkdownToken, block_copy_token)
            block_copy_token_leading_spaces = block_copy_list_token.leading_spaces
            are_same = (
                found_markdown_token_leading_spaces == block_copy_token_leading_spaces
            )
            assert not are_same
            POGGER.debug(
                "__handle_blank_line_bravo>>found_markdown_list_token>>$",
                found_markdown_list_token,
            )
            found_markdown_list_token.remove_last_leading_space()
            POGGER.debug(
                "__handle_blank_line_bravo>>found_markdown_list_token>>$",
                found_markdown_list_token,
            )

    @staticmethod
    def __handle_blank_line_token_stack(
        parser_state: ParserState,
    ) -> Tuple[Optional[List[MarkdownToken]], bool, Optional[RequeueLineInfo]]:
        (
            is_processing_list,
            in_index,
        ) = LeafBlockProcessorParagraph.check_for_list_in_process(parser_state)
        # POGGER.debug(
        #     "hbl>>is_processing_list>>$>>in_index>>$>>last_stack>>$",
        #     is_processing_list,
        #     in_index,
        #     parser_state.token_stack[-1],
        # )

        requeue_line_info: Optional[RequeueLineInfo] = None
        new_tokens: Optional[List[MarkdownToken]] = None
        force_default_handling = False
        last_stack_token = None

        force_default_handling, last_stack_token, requeue_line_info, new_tokens = (
            TokenizedMarkdown.__handle_blank_line_token_stack_multiline_block(
                parser_state
            )
        )
        if not force_default_handling:
            if parser_state.token_stack[-1].is_code_block:
                if parser_state.count_of_block_quotes_on_stack():
                    POGGER.debug("hbl>>code block within block quote")
                else:
                    POGGER.debug("hbl>>code block")
                    new_tokens = []
            elif parser_state.token_stack[-1].is_html_block:
                POGGER.debug("hbl>>check_blank_html_block_end")
                new_tokens = HtmlHelper.check_blank_html_block_end(parser_state)
                POGGER.debug("hbl<<check_blank_html_block_end")
            elif (
                is_processing_list
                and parser_state.token_document[-1].is_blank_line
                and parser_state.token_document[-2].is_list_start
            ):
                POGGER.debug("hbl>>double blank in list")
                new_tokens, _ = TokenizedMarkdown.__close_open_blocks(
                    parser_state, until_this_index=in_index, include_lists=True
                )
                POGGER.debug("hbl<<double blank in list")

        if requeue_line_info and last_stack_token is not None:
            parser_state.abc(requeue_line_info, last_stack_token)

        return new_tokens, force_default_handling, requeue_line_info

    @staticmethod
    def __handle_blank_line_token_stack_multiline_block(
        parser_state: ParserState,
    ) -> Tuple[
        bool,
        Optional[StackToken],
        Optional[RequeueLineInfo],
        Optional[List[MarkdownToken]],
    ]:

        force_default_handling = False
        last_stack_token = None
        requeue_line_info: Optional[RequeueLineInfo] = None
        new_tokens: Optional[List[MarkdownToken]] = None

        if parser_state.token_stack[-1].was_link_definition_started:
            force_default_handling = True
            last_stack_token = parser_state.token_stack[-1]
            POGGER.debug(
                "hbl>>process_link_reference_definition>>stopping link definition"
            )
            empty_position_marker = PositionMarker(-1, 0, "")
            assert parser_state.original_line_to_parse is not None
            (
                _,
                _,
                did_pause_lrd,
                requeue_line_info,
                new_tokens,
            ) = LinkReferenceDefinitionHelper.process_link_reference_definition(
                parser_state,
                empty_position_marker,
                "",
                "",
                "",
                0,
                0,
                parser_state.original_line_to_parse,
            )
            assert not did_pause_lrd, "LRD parsing must not be paused."
            POGGER.debug(
                "hbl<<process_link_reference_definition>>stopping link definition"
            )
        elif parser_state.token_stack[-1].was_table_block_started:
            force_default_handling = True
            last_stack_token = parser_state.token_stack[-1]
            POGGER.debug("hbl>>process_table>>stopping table")
            unmod = parser_state.original_line_to_parse
            assert unmod is not None
            empty_position_marker = PositionMarker(-1, 0, "")
            (
                _,
                _,
                did_pause_lrd,
                requeue_line_info,
                new_tokens,
            ) = TableBlockHelper.process_table_rows(
                parser_state, empty_position_marker, "", "", unmod, 0, 0, ""
            )
            assert not did_pause_lrd, "Table parsing must not be paused."
            POGGER.debug("hbl<<process_table>>stopping table")
        return force_default_handling, last_stack_token, requeue_line_info, new_tokens

    @staticmethod
    def __handle_blank_line_close_first_quote(
        parser_state: ParserState, new_tokens: Optional[List[MarkdownToken]]
    ) -> Optional[List[MarkdownToken]]:

        stack_index = 1
        while (
            stack_index < len(parser_state.token_stack)
            and not parser_state.token_stack[stack_index].is_block_quote
        ):
            stack_index += 1

        close_tokens, _ = parser_state.close_open_blocks_fn(
            parser_state,
            include_block_quotes=True,
            include_lists=True,
            until_this_index=stack_index,
            was_forced=True,
        )
        POGGER.debug("close_tokens>>$", close_tokens)
        if new_tokens is not None:
            new_tokens.extend(close_tokens)
        elif close_tokens:
            new_tokens = close_tokens
        return new_tokens

    @staticmethod
    def __handle_blank_line_in_block_quote(
        parser_state: ParserState, new_tokens: Optional[List[MarkdownToken]]
    ) -> None:
        # assert not new_tokens or (new_tokens and not new_tokens[-1].is_link_reference_definition)
        # assert new_tokens is None or len(new_tokens) <= 1
        if new_tokens and len(new_tokens) == 1 and new_tokens[0].is_list_end:
            return
        stack_index = parser_state.find_last_container_on_stack()
        if parser_state.token_stack[stack_index].is_list:
            return
        POGGER.debug_with_visible_whitespace("new_tokens>>$", new_tokens)
        POGGER.debug("stack_index>>$", stack_index)
        if new_tokens and stack_index > 0:
            POGGER.debug_with_visible_whitespace("new_tokens[-1]>>$", new_tokens[-1])
            assert (
                new_tokens[-1].is_html_block_end
                or new_tokens[-1].is_link_reference_definition
                or new_tokens[-1].is_table_end
            ) and stack_index == (
                len(parser_state.token_stack) - 1
            ), "If here, must be in the right context."
            close_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                only_these_blocks=[BlockQuoteStackToken],
                include_block_quotes=True,
            )
            POGGER.debug("close_tokens>>$", close_tokens)
            new_tokens.extend(close_tokens)

        stack_index = parser_state.find_last_container_on_stack()
        POGGER.debug(
            "blank>>bq_start>>$",
            parser_state.token_stack[stack_index],
        )
        if stack_index > 0 and parser_state.token_stack[stack_index].is_block_quote:
            block_quote_token = cast(
                BlockQuoteMarkdownToken,
                parser_state.token_stack[stack_index].matching_markdown_token,
            )
            POGGER.debug("hblibq>>last_block_token>>$", block_quote_token)
            POGGER.debug(
                "hblibq>>leading_text_index>>$", block_quote_token.leading_text_index
            )
            block_quote_token.add_bleading_spaces("")
            POGGER.debug("hblibq>>last_block_token>>$", block_quote_token)
            POGGER.debug(
                "hblibq>>leading_text_index>>$", block_quote_token.leading_text_index
            )

    def __process_front_matter_header_if_present(
        self,
        first_line_in_document: Optional[str],
        line_number: int,
        requeue: List[str],
    ) -> Tuple[Optional[str], int, List[str]]:
        assert self.__source_provider, "Source provider must be defined by this point."
        assert (
            self.__parse_properties
        ), "Parse properties must be defined by this point."
        POGGER.debug(
            "is_front_matter_enabled>>$",
            self.__parse_properties.is_front_matter_enabled,
        )
        if (
            first_line_in_document is not None
            and self.__parse_properties.is_front_matter_enabled
        ):
            assert (
                self.__extension_manager is not None
            ), "Extension manager must be initialized by this point."
            raw_extension = self.__extension_manager.get_extension_instance(
                FrontMatterExtension().get_identifier()
            )
            front_matter_extension = cast(FrontMatterExtension, raw_extension)
            (
                first_line_in_document,
                line_number,
                requeue,
            ) = front_matter_extension.process_header_if_present(
                first_line_in_document,
                line_number,
                requeue,
                self.__source_provider,
                self.__tokenized_document,
            )
        return first_line_in_document, line_number, requeue
